/*
 * $Id: main.cpp,v 1.26 2001/12/25 16:40:37 amirs Exp $
 */
/* This is the main entry point into the system call tracker 
 * configuration utility
 * @author: mulix@actcom.co.il
 */

#include <iostream>
#include <fstream>
#include <strstream>
#include <list>

extern"C"{
#include <unistd.h>
#include <errno.h>
#include <string.h>
    //for stat()
#include <sys/stat.h>
    //for uid_t
#include <sys/types.h>
}

#include "sct_ctrl_lib/sct_ctrl_lib.h"
#include "rule.h"
#include "logger.h"
#include "parser.h"
#include "syscalls.h"
#include "tokens_to_types.h"

const char DEFAULT_CONFIG_FILE_NAME[] = "sct_example.conf";
const char DEFAULT_SYSCALLS_INFO_LOCATION[] = "../sct_rules_module/syscalls.dat";

/* forward declarations */
bool load_config_file(const char * config_file_name,
		      strstream& data);
bool parse_config_file(strstream& data,
		       string& log_format_before, string& log_format_after,
		       list<rule>& rules, 
		       const string& config_file_name);

#ifdef SCT_CONFIG_DEBUG
int debug = 1;
#else /* SCT_CONFIG_DEBUG is not defined */
int debug = 0;
#endif

void usage(const char* name)
{
    cerr<<name<<": a configuration and control utility for the syscall tracker"<<endl;
    cerr<<"Usage:\t"<<name<<" [command] [file name]"<<endl;
    cerr<<"\t"<<name<<" print: print all currently registered rules"<<endl;
    cerr<<"\t"<<name<<" delete: delete all currently registered rules"<<endl;
    cerr<<"\t"<<name<<" upload [file name]: parse configuration file\n"
	<<"\t\tand upload rules to the kernel module"<<endl;
    cerr<<"\t"<<name<<" check [file name]: parse and print all rules\n"
	<<"\t\tin the configuration file, but don't upload them to the\n"
	<<"\t\tkernel module"<<endl;
    exit(1);
}

void parse_options(char** argv, int argc, bool& print, 
		   bool& del, bool& check, bool& upload, 
		   bool& verbose,
		   const char*& fname)
{
    print = del = check = upload = verbose = false;
    char* cmd = argv[1];
    if (!strcmp(cmd, "--verbose")){
        verbose = true;
	argv++;
	argc--;
        cmd = argv[1];
    }
    if (!strcmp(cmd, "print")){
	print = true; 
    } else if (!strcmp(cmd, "delete")){
	del = true; 
    } else if (!strcmp(cmd, "upload")){
	upload = true;
    } else if (!strcmp(cmd, "check")){
	check = true;
    } else 
	usage(argv[0]);

    if (upload || check){
	fname = ((argc == 3) ? argv[2] : DEFAULT_CONFIG_FILE_NAME);
    }
}

static int 
print_module_rules(sct_ctrl_t ctrl)
{
    int rc = sct_ctrl_print_all_rules(ctrl);
    if (!rc){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "Error: failed printing all rules.\n");
	logger::log(logger::LOG_LEVEL_ERROR,
		    "       print all rules returned '%d'\n",
		    rc);
    }

    return rc; 
}

static int 
delete_all_rules(sct_ctrl_t ctrl)
{
    int rc = sct_ctrl_del_all_rules(ctrl);
    if (!rc){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "Error: failed deleting all previous module rules.\n");
	logger::log(logger::LOG_LEVEL_ERROR,
		    "       delete all rules returned '%d'\n",
		    rc);
    }

    return rc;
}

int main(int argc, char ** argv)
{
    bool print, erase, check, upload, verbose;
    const char* config_file_name; 
    int rc;

    if ((argc < 2) || (argc > 3))
	usage(argv[0]);

    parse_options(argv, argc, print, erase, check,
		  upload, verbose, config_file_name);

    if (verbose)
	logger::set_verbosity_level(logger::LOG_LEVEL_INFO);

    /* init the control library */
    sct_ctrl_t sct_ctrl;
    if (!check){
	sct_ctrl = sct_ctrl_init();
	if (!sct_ctrl){
	    logger::log(logger::LOG_LEVEL_ERROR,
			"failed to initialize the ctrl library\n");
	    return 1;
	}
    }
     
    if (print) {
	rc = print_module_rules(sct_ctrl);
	if (rc)
	    cout << "module rules printed to system log file" << endl;
	return (!rc);
    }

    if (erase) {
	rc = delete_all_rules(sct_ctrl); 
	if (rc)
	    cout << "all module rules were deleted sucessfully" << endl;
	return (!rc);
    }

    /* it's either check or upload. let's go on */

    strstream data;
    if (!strcmp(config_file_name, "-")) {
	cin>>data.rdbuf();
    } else if (!load_config_file(config_file_name, data)){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "failed to read the config file '%s'. quitting.\n", 
		    config_file_name);
	return 1;
    }
    //the list of rules
    list<rule> rules;
    //the log format
    string log_format_before = "";
    string log_format_after = "";

    if (!parse_config_file(data,
			   log_format_before,
			   log_format_after,
			   rules,
			   config_file_name)){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "failed to parse the config file. quitting.\n");
	return 1;
    }

    if (check || upload) {
	if (log_format_before != "") {
	    cout << "parsed log format 'before': '" << log_format_before
		<< "'" << endl;
	}
	if (log_format_after != "") {
	    cout << "parsed log format 'after': '" << log_format_after
		<< "'" << endl;
	}
	cout << "parsed rule list:" << endl;
	list<rule>::const_iterator it = rules.begin();
	for (; it != rules.end(); ++it){
	    cout << *it << endl;
	}
	if (check)
	    return 0;
    }

    /* make sure 'log_format' contains a new-line, unless its empty. */
    if (log_format_before != "") {
	log_format_before += "\n";
    }
    if (log_format_after != "") {
	log_format_after += "\n";
    }

    rc = delete_all_rules(sct_ctrl);
    if (!rc) {
	return 1;
    }

    list<rule>::const_iterator it = rules.begin();
    for (; it != rules.end(); ++it){
	strstream rule_data;
	rule_data << *it << '\0';
	logger::log(logger::LOG_LEVEL_INFO,
		    "handling rule:\n%s",
		    rule_data.str());
	filter_node_t* filter = (*it).filter_node();
	action_t* act = ((*it).get_rule_when() == RW_BEFORE
				? (*it).get_action_t(log_format_before)
				: (*it).get_action_t(log_format_after));
	rc = sct_ctrl_add_rule(sct_ctrl, (*it).get_rule_id(), 
			       (*it).get_rule_when(),
			       (*it).get_syscall_identifier(), 
			       filter, act, 1);
	if (!rc){
	    logger::log(logger::LOG_LEVEL_ERROR,
			"Error: failed adding rule '%s'\n",
			(*it).get_rule_name().c_str());
	    logger::log(logger::LOG_LEVEL_ERROR,
			"       sct_ctrl_add_rule returned %d. quitting!\n",
			rc);
	    return -1;
	}
	free_action(act);
    }
    cout << "successfully uploaded rules from file '"
	 << config_file_name << "'" << endl;
     
    return 0;
}


bool load_config_file(const char * config_file_name,
		      strstream& data)
{
    //first of all, make some sanity checks on the file

    //preferably, that user will be root, but we relax that requirement for
    //development
    struct stat stat_info;
    if (stat(config_file_name, &stat_info)==-1){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "load_config_file: stat failed, with error '%s'\n",
		    strerror(errno));
	return false;
    }

    //the config file is a regular file
    if (!S_ISREG(stat_info.st_mode)){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "load_config_file: file '%s' is not a regular file\n", 
		    config_file_name);
	return false;
    }
#if 0
    //the config file is owned by the current user
    if (stat_info.st_uid != geteuid()){
	logger::log(logger::LOG_LEVEL_ERROR,
		    "load_config_file: file is not owned by the current euid\n");
	return false;
    }

    //the config file is only writeable by the current user. 
    if ((stat_info.st_mode & (S_IRUSR | S_IWUSR)) != (S_IRUSR & S_IWUSR)){
	log("load_config_file: wrong permissions on config_file");
	log(config_file_name);
	return false;
    }
#endif

    //now, that we've verified the file is usable by us, let's load
    //its contents into a buffer.
    ifstream in(config_file_name);
    in>>data.rdbuf();
    return true;
}

bool 
parse_config_file(strstream& data,
		  string& log_format_before, string& log_format_after,
		  list<rule>& rules,
		  const string& config_file_name)
{
    /* maps from string to enumerated types     */
    /* must be done here since the syscalls.dat */
    /* parser uses them as well                 */
    tokens::init();
    
    // load the syscall information file
    if (!syscalls::load(DEFAULT_SYSCALLS_INFO_LOCATION)) {
	logger::log(logger::LOG_LEVEL_ERROR,
		    "Error: failed to find syscall information "
		    "file at '%s'.\n", DEFAULT_SYSCALLS_INFO_LOCATION);
	return false;
    }
    
    //create a new parser
    parser p(data, config_file_name);
    //start parsing the config data
    return p.parse(log_format_before, log_format_after, rules);
}
