/*
 *$Id: parser.cpp,v 1.29 2001/12/16 22:29:16 mulix Exp $
 */
/* parser implementation */
#include <string>
#include <strstream>

#include "parser.h"
#include "parse_exception.h"
#include "token.h"
#include "validity_checks.h"
#include "logger.h"
#include "param_types.h"
#include "tokens_to_types.h"
#include "filter_parser.h"

#define WHITESPACE " \t\n\r"

parser::parser(strstream& config_data, string file_name):
    _data(config_data), _fname(file_name)
{
}


/* the main entry point. performs initial processing and then parses
   each rule */
bool parser::parse(string& log_format_before,
		   string& log_format_after,
		   list<rule>& rules)
{
    //parsing the configuration data.

    /*strip comments from the file. we do this in a pretty bizzare way
      due to problems encountered with strstreams. what we need to do here 
      is read all of the data from the strstream, get rid of all comment lines
      and then spill it into a new strstream and replace the old one. */
    string data;
    char buffer[2048];
    bool multi_line_comment = false;
     
    //read the data line by line from the file. 
    //if the line is a comment dump it. 
    while (_data.getline(buffer, sizeof(buffer))){
	// strip c style multi line comments
	char* comment_start;
	while (multi_line_comment || (comment_start = strstr(buffer, "/*"))){
	    if (!comment_start)
		comment_start = buffer; 

	    char* comment_end = strstr(buffer, "*/");
	    if (comment_end){
		/* get rid of the comment by copying what's beyond it over it */
		memcpy(comment_start, comment_end + 2, 
		       buffer + sizeof(buffer) - comment_end);
		multi_line_comment = false;
	    } else {
		multi_line_comment = true;
		break;
	    }
	}
	/* if this is a comment, we do an ugly hack and push in 
	   an empty line to get the line numbers in diagnostics 
	   correctly. not the best way to do it, but it's simple */
	if ((multi_line_comment) ||
	    (buffer[0] == '#') || 
	    ((buffer[0] == '/') && (buffer[1] == '/'))){
	    data+='\n';
	    continue;
	}

	/* add the current buffer to the data */
	data+=buffer;

	/* getline discarded the '\n' */
	data+='\n';
    }
     
    //do the actual parsing
    int line_num = 1;
    token t;
    string::size_type location = 0;
    //read_token will return a token with the entire rule data contained
    //and a type of RULE
    try {
	while (read_token(t, data, &location, &line_num)){
	    if (t.get_type() == PR_RULE){
		rule r(*this);
		r.set_rule_line_num(t.get_type_line_num());
		r.set_rule_data_line_num(t.get_value_line_num());
		parse_rule(r, t.get_value());
		rules.push_back(r);
	    }else if (t.get_type() == PR_LOG_FORMAT){
		parse_log_format(t.get_value_line_num(), t.get_value(), 
				 log_format_before, log_format_after);
	    }else{ //error in the config file. log it. 
		throw parse_exception_t(get_file_name(), line_num, 
					"error in configuration file", t);
	    }
	}
    }catch (exception_intf& e){
	log_error(e);
	return false;
    }
    return true;
}

void
parser::parse_rule(rule& r, const string& rule)
{
    token t;
    string::size_type pos = 0;
    string filter_clause;
    bool has_filter_clause = false; 
    int type_line_num = 0; 
    int value_line_num = 0;
    int line_num = r.get_rule_data_line_num();
    while(read_token(t, rule , &pos, &line_num)){
	//read tokens from the data and act based on their types. 
	switch (t.get_type()){
	case(PR_SYSCALL_NAME):{
	    r.set_syscall_identifier(t.get_value());
	    r.set_syscall_name(t.get_value());
	    break;
	}
	case(PR_RULE_NAME):{
	    r.set_rule_name(t.get_value());
	    break;
	}
	case(PR_RULE_WHEN):{
	    if (t.get_value().compare("before") == 0){
	        r.set_rule_when(RW_BEFORE);
	    }
	    else if (t.get_value().compare("after") == 0){
	        r.set_rule_when(RW_AFTER);
	    }
	    else {
		string err = "unknown 'when' value '" + t.get_value() +
			     "' in configuration file";
		throw parse_exception(get_file_name(), line_num, err);
	    }
	    break;
	}
	case(PR_PARAM_CONDITION):{
	    param_conditional* pcond = r.new_param_conditional();
	    if (!pcond){
		throw parse_exception_t(get_file_name(), line_num, 
				      "out of memory", t);
	    }
	    pcond->set_line_num(t.get_type_line_num());
	    pcond->set_data_line_num(t.get_value_line_num());
	    parse_param_condition(pcond, t.get_value());
	    break; 
	}
	case(PR_PROCESS_CONDITION):{
	    process_conditional* pcond = r.new_process_conditional();
	    if (!pcond){
		throw parse_exception_t(get_file_name(), line_num, 
				      "out of memory", t);
	    }
	    pcond->set_line_num(t.get_type_line_num());
	    pcond->set_data_line_num(t.get_value_line_num());
	    parse_process_condition(pcond, t.get_value());
	    break;
	}
	case(PR_ACTION):{
	    r.set_action(t.get_value());
	    break;
	}
	case(PR_CONDITIONS):{
	    r.set_condition_matching(t.get_value());
	    break;
	}
	case (PR_FILTER_EXPRESSION):{
	    /* we defer parsing the filter expression until we parse 
	       all of the other rule fields, in particular, the 
	       syscall name */
	    if (has_filter_clause)
		throw parse_exception_t(get_file_name(), line_num, 
					"more than one filter expression in file", t);
	    has_filter_clause = true;
	    filter_clause = t.get_value();
	    type_line_num = t.get_type_line_num();
	    value_line_num = t.get_value_line_num();
	    break;
	}
	default:
	    throw parse_exception_t(get_file_name(), line_num, 
				    "invalid token parsed", t);
	}
    }

    /* if we have the filter parser clause to parse, do it now */
    if (has_filter_clause) {
	filter_parser p(*this, r.get_syscall_name());
	p.set_line_num(type_line_num);
	p.set_data_line_num(value_line_num);
	filter_node_t* exp = NULL;
	p.parse(exp, filter_clause);
	r.set_filter_expression(exp);
    }
    
    //finalize the rule
    (void)r.finalize();

    //validate the rule, by making sure it has all required fields and 
    //no collisions in the instructions. 
    (void)r.validate();
}

void parser::parse_log_format(int line_num, const string& log_formats, 
			      string& log_format_before, 
			      string& log_format_after)
{
    token t;
    string::size_type pos = 0;
    string lf_before, lf_after;
    /* warn if the user sets a 'before' or */
    /* 'after' format, *and* a general format. */
    const unsigned int got_nothing = 0x0; 
    const unsigned int got_before = 0x1;
    const unsigned int got_after = 0x2;
    const unsigned int got_default = 0x4;
    unsigned int got = 0;
    while(read_token(t, log_formats , &pos, &line_num)){
	switch (t.get_type()){
	case(PR_LOG_FORMAT_BEFORE): {
	    lf_before = t.get_value();
	    got |= got_before;
	    break;
	}
	case(PR_LOG_FORMAT_AFTER): {
	    lf_after= t.get_value();
	    got |= got_after;
	    break;
	}
	case(PR_LOG_FORMAT_DEFAULT): {
	    /* the user must have specified the log_format  */
	    /* value directly. set before here, and we'll */
	    /* set after later */
	    lf_before = lf_after = t.get_value();
	    got |= got_default;
	    break;
	}
	default: {
	    got = got_nothing;
	    break;
	}
	}
    }
    /* what did the user specify? */
    switch (got){
    case (got_nothing): {
	strstream buf;
	buf << "log_format directive with an invalid log format" << endl << ends;
	throw validate_exception(get_file_name(), line_num, buf.str());
	/* not reached */
	break;
    }
    case (got_before): /* fallthrough */
    case (got_after): {
	strstream buf;
	buf << "only '" << ((got & got_before) ? "before" : "after")
	    << "' log_format directive found. you must specify '"
	    << ((got & got_before) ? "after" : "before" )
	    << "' as well" << endl << ends;
	throw validate_exception(get_file_name(), line_num, buf.str());
	/* not reached */
	break;
    }
    case ((got_before | got_after)): {
	/* we should validate it here */
	log_format_before = lf_before;
	log_format_after = lf_after;
	break;
    }
    case (got_default): {
	/* validate, validate, validate */
	log_format_before = log_format_after = lf_before;
	break;
    }
    default: { /* any other combination of default and before/after */
	strstream buf;
	buf << "log_format 'before' or 'after' specified together with "
	    "a default log_format." << endl << ends;
	throw validate_exception(get_file_name(), line_num, buf.str());
	break;
    }
    }
}

bool parser::read_token(token& t,
			const string& rule,
			string::size_type* pos,
			int* p_line_num)
{
    int token_type_line_num;
    int token_value_line_num;
    //a complex token is one which has the begin and end ('{' and '}') chars
    //after the name. 
    bool complex = false;
    int complex_line_num;

    //find the first non whitespace character.
    string::size_type beg = rule.find_first_not_of(WHITESPACE, *pos);
    if (beg == rule.npos){
	return false;
    }

    // update line number.
    *p_line_num += count_lines(rule, *pos, beg);
    token_type_line_num = *p_line_num;
     
    //find the end of this word. this is either whitespace, the '=' sign 
    //or the '{' sign. 
    string::size_type end = rule.find_first_of(WHITESPACE":={", beg+1);
    if (end == rule.npos){
	return false;
    }

     
    //is it a complex token?
    if (rule[end] == '{')
	complex = true;
     
    //this word is the 'type' of the token. 
    //we take the address of a temp variable here, so care must be taken
    //not to use it outside of this function.
    const string& tok = rule.substr(beg, end - beg);
//     cerr<<"***token: "<<tok;

    // update line number.
    *p_line_num += count_lines(rule, beg+1, end);

    //now find the beginning of the 'value' part of the token
    beg = rule.find_first_not_of(WHITESPACE, end+1);
    if (beg == rule.npos){
	return false;
    }
     
    // update line number.
    *p_line_num += count_lines(rule, end+1, beg);

    //again, check if this token is complex. 
    if (rule[beg] == '{'){
	complex = true;
	complex_line_num = *p_line_num;
	//if it is complex, skip forward a few chars until the 
	//actual beginning of the value. 
	string::size_type old_beg = beg;
	beg = rule.find_first_not_of(WHITESPACE,beg+1);
	if (beg == rule.npos)
	    return false;
        // update line number.
        *p_line_num += count_lines(rule, old_beg+1, beg);
    }

    token_value_line_num = *p_line_num;

    //if it's not a complex token, the value is only one word, so 
    //it ends with whitespace. 
    if (!complex){
	end = rule.find_first_of(WHITESPACE, beg+1);
	if (end == rule.npos)
	    return false;
        // update line number.
        *p_line_num += count_lines(rule, beg+1, end);
    }else{
	//if it is a complex token, it might contain other complex
	//tokens, so we need to find the '}' sign matching ours. 
	int count = 1;
	end = rule.find_first_of("{}", beg+1);
	while (end != rule.npos){
	    if (rule.at(end) == '{'){
		++count;
	    }else{
		--count;
	    }
	    if (count == 0)
		break;
	    end = rule.find_first_of("{}", end+1);
	}
	if (count != 0){
	    throw parse_exception(get_file_name(), complex_line_num,
				  "mismatched brackets in input");
	}
        // update line number.
        *p_line_num += count_lines(rule, beg+1, end);
    }

    parser_token_t type = tokens::get_parser_token(tok);
    t.set_type(type);
    t.set_type_line_num(token_type_line_num);
    
    if (t.get_type() == PR_LAST_TOKEN_TYPE) { 
	string err = "unknown string '" + tok + "' in configuration file";
	throw parse_exception(get_file_name(), token_type_line_num, err);
    }

    //ok, copy the value part. 
    const string& tok_value = rule.substr(beg, end - beg);
    t.set_value(tok_value);
    t.set_value_line_num(token_value_line_num);

    *pos = end+1;
    return true;
}


void parser::parse_param_condition(
    param_conditional* pcond, const string& data)
{
    token t;
    string::size_type pos = 0;
    int line_num = pcond->get_data_line_num();
    
    while (read_token(t, data, &pos, &line_num)){
	switch(t.get_type()){
	case(PR_PARAM_LOCATION):{
	    pcond->set_location(t.get_value());
	    break;
	}
	case(PR_PARAM_LOCATION_FIELD):{
	    pcond->set_location_field(t.get_value());
	    break;
	}
	case(PR_OPERATOR):{
	    pcond->set_operator(t.get_value());
	    break;
	}
	case(PR_VALUE):{
	    pcond->set_value(t.get_value());
	    break;
	}
	default:
	    throw parse_exception_t(get_file_name(), line_num, 
				  "unknown token in parameter condition", t);
	}
    }
}

void parser::parse_process_condition(
    process_conditional* pcond, const string& data)
{
    token t;
    string::size_type pos = 0;
    int line_num = pcond->get_data_line_num();

    while (read_token(t, data, &pos, &line_num)){
	switch(t.get_type()){
	case(PR_PROCESS_FIELD):{
	    pcond->set_process_field(t.get_value());
	    break;
	}
	case(PR_OPERATOR):{
	    pcond->set_operator(t.get_value());
	    break;
	}
	case(PR_VALUE):{
	    pcond->set_value(t.get_value());
	    break;
	}
	default:
	    throw parse_exception_t(get_file_name(), line_num, 
				    "unknown token in process condition", t);
	}
    }
}

void parser::log_error(const exception_intf& e)
{
    strstream buf; 
    e.print(buf);
    buf << endl << ends;
    
    logger::log(logger::LOG_LEVEL_ERROR, "%s", buf.str());
}

// count the number of lines (new-line chars)  in the given string,
// between 'beg' and 'end'.
int parser::count_lines(const string& data,
			 string::size_type beg,
			 string::size_type end)
{
    int line_span = 0;
    string::size_type curr = data.find_first_of("\n", beg);
    if (end > data.length())
	end = data.length();

    while (curr <= end) {
	line_span++;
        curr = data.find_first_of("\n", curr+1);
    }

    return line_span;
}
