/*
 * $Id: parser.h,v 1.11 2001/12/16 22:29:16 mulix Exp $
 */
/*  parser is responsible of parsing the configuration file, as defined in 
 *  the config.file document. it then builds a list of rules which can
 *  sent to the module
 *
 *  @author: mulix@actcom.co.il
 */

#ifndef _PARSER_H_
#define _PARSER_H_

#include <string>
#include <strstream>

#include "token.h"
#include "rule.h"
#include "process_conditional.h"
#include "param_conditional.h"
#include "parse_exception.h"

using namespace std;

class parser
{
 public:
    /* constructor.
       @param: the configuration data, passed as a strstream  */
    parser(strstream& config_data, string file_name);

    /* parse is the one and only entry point into the parser. 
       if the parsing failed the function will return it. 
       if the parsing of any rule is successful, we return the 
       parsed rules in the rules list. 
       returns true if parsing of ALL rules was successful.    */
    bool parse(string& log_format_before,
	       string& log_format_after,
	       list<rule>& rules);

    inline const string& get_file_name() const { return _fname; } 

    /* count the number of lines (new-line chars)  in the given string,
       between 'beg' and 'end'. */
    int count_lines(const string& data,
		    string::size_type beg,
		    string::size_type end);
    
 private:
    /* parse_rule accepts a string of rule data and parses it into a 
       rule structure. */
    void parse_rule(rule& r, const string& data);

    /* parse_log_format accepts a string containing log_format information,
       parses it and stores it in log_format_before, log_format_after or both, 
       as applicable. if an error occurs, an exception is thrown. */
    void parse_log_format(int line_num, const string& log_formats, 
			  string& log_format_before, 
			  string& log_format_after);

    /* read_token does most of the parsing. it accepts a string of rule
       data, a position to start parsing from, and a line number to
       start counting from.
       it returns in its arguments a token class with data about the parsed 
       token, the position to start the next parsing from, and the line number
       of the end of the token.
       returns false when there are no more tokens to be read
       throws a parse_exception if encounters an unknown token
    */
    bool read_token(token& t,
		    const string& rule,
		    string::size_type* pos,
		    int* p_line_num);

    /* parse_process_condition accepts a process_conditional structure 
       and a string of rule data. it parses the data and fills the 
       process conditional. */
    void parse_process_condition(process_conditional* pcond, const string& data);

    /* parse_process_condition accepts a param_conditional structure 
       and a string of rule data. it parses the rule data and fills
       the param_conditional structure. */
    void parse_param_condition(param_conditional* pcond, const string& data);

    /* log_error accepts a parse_exception and logs it to the log file */
    void log_error(const exception_intf& e);
     
    /* members */
    strstream& _data;

    string _fname;
    unsigned int _line;
};

#endif
