/*
 * $Id: dev_file_impl.c,v 1.4 2001/12/09 19:55:06 chooga Exp $
 */ 

/* this file implements the device file communication protocol between */
/* the kernel module and userspace. for protocol details, see          */
/* doc/sct_device_control_file_design.txt                              */ 

#include <assert.h>
#include <stdio.h>

#include "sct_ctrl_lib.h"
#include "dev_file_impl.h"

/* forward declaration */
static int
dev_file_common(sct_ctrl_t ctrl, unsigned int cmd,
		int rule_id, rule_when_t when, int syscall_id, 
		filter_node_t* filter, action_t* act, 
		int enable_rule);

/* add a rule with the given rule id to the given syscall,    */
/* where the rule is made of this filter and this action      */
/* and 'enable rule' specifies if the rule should be          */
/* immediately enabled or not                                 */
/* return 0 on failure, non 0 on success                      */
int 
dev_file_add_rule(sct_ctrl_t ctrl, int rule_id, rule_when_t when,
		  int syscall_id, 
		  filter_node_t* filter, action_t* act, 
		  int enable_rule)
{
    return dev_file_common(ctrl, CTL_TRACKER_CMD_ADD_RULE, 
			   rule_id, when, syscall_id, 
			   filter, act, enable_rule);
}

/* delete a rule with the given rule id to the given syscall, */
/* return 0 on failure, non 0 on success                      */
int
dev_file_del_rule(sct_ctrl_t ctrl,
		  int rule_id,
		  rule_when_t when,
		  int syscall_id)
{
    return dev_file_common(ctrl, CTL_TRACKER_CMD_DEL_RULE,
			   rule_id, when, syscall_id, 
			   NULL /* filter */, NULL /* act */,
			   0 /* enable_rule */);
}

/* enable a rule with the given rule id to the given syscall, */
/* return 0 on failure, non 0 on success                      */
int 
dev_file_enable_rule(sct_ctrl_t ctrl,
		     int rule_id,
		     rule_when_t when,
		     int syscall_id)
{
    return dev_file_common(ctrl, CTL_TRACKER_CMD_ENABLE_RULE, 
			   rule_id, when, syscall_id, 
			   NULL /* filter*/, NULL /* act */,
			   0 /* enable_rule */);
}

/* disable a rule with the given rule id to the given syscall, */
/* return 0 on failure, non 0 on success                       */
int 
dev_file_disable_rule(sct_ctrl_t ctrl,
		      int rule_id,
		      rule_when_t when,
		      int syscall_id)
{ 
    return dev_file_common(ctrl, CTL_TRACKER_CMD_DISABLE_RULE,
			   rule_id, when, syscall_id, 
			   NULL /* filter*/, NULL /* act */,
			   0 /* enable_rule */);
}

/* delete all rules for this syscall, return 0 on failure, non */
/* 0 on succes                                                 */
int 
dev_file_del_all_syscall_rules(sct_ctrl_t ctrl, int syscall_id)
{
    return dev_file_common(ctrl, 
			   CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL,
			   -1 /* rule_id */, 
			   RW_BEFORE /* when */, 
			   syscall_id, 
			   NULL /* filter*/, NULL /* act */,
			   0 /* enable_rule */);
}
/* delete all rules for all syscalls, return 0 on failure, non */
/* 0 on succes                                                 */
int 
dev_file_del_all_rules(sct_ctrl_t ctrl)
{
    return dev_file_common(ctrl, 
			   CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES,
			   -1 /* rule_id */, 
			   RW_BEFORE /* when */, 
			   -1 /* syscall_id */, 
			   NULL /* filter*/, NULL /* act */,
			   0 /* enable_rule */);
}

/* print the current rules, return 0 on failure, non */
/* 0 on succes                                       */
int
dev_file_print_all_rules(sct_ctrl_t ctrl)
{
    return dev_file_common(ctrl, 
			   CTL_TRACKER_CMD_PRINT_ALL_RULES,
			   -1 /* rule_id */, 
			   RW_BEFORE /* when */, 
			   -1 /* syscall_id */, 
			   NULL /* filter*/, NULL /* act */,
			   0 /* enable_rule */);
}


/* the common dev file access routine. we return 0 on failure, */
/* non 0 on success                                            */
int
dev_file_common(sct_ctrl_t ctrl, unsigned int cmd,
		int rule_id, rule_when_t when, int syscall_id, 
		filter_node_t* filter, action_t* act, 
		int enable_rule)
{
    /* build the rule, and pass it on the the protocol layer */

    int rc; 
    int serial_len = 0;
    unsigned char buf[1000];
    unsigned char* p_buf = NULL;
    
    assert(rule_id >= 0 && syscall_id >= 0);
    if ((rule_id < 0) ||(syscall_id < 0))
	return -1;
    
    if (cmd == CTL_TRACKER_CMD_ADD_RULE) {
	serial_len = 
	    serialize_tracking_rule(rule_id, when, syscall_id, act, 
				    filter, buf, sizeof(buf));
	
	if (serial_len <= 0){
	    fprintf(stderr, "failed to serialize tracking rule, "
		    "error '%d'\n", serial_len);
	    return -1;
	}
	p_buf = buf;
    } 
    
    rc = dev_file_do_protocol(ctrl, cmd, rule_id, when, syscall_id, 
			      p_buf, serial_len, enable_rule);
    
    return (rc == CTL_TRACKER_SUCCESS);
}
