#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#include "sct_ctrl_lib.h"
#include "sct_ctrl_lib_private.h"
#include "sct_rules_module/module_interface.h"

typedef struct dev_request_type
{
    unsigned int size;
    unsigned int id; 
    unsigned int type;
    int rule_id;
    rule_when_t when;
    int syscall_id;
    unsigned int tracking_rule_len;
    unsigned char* tracking_rule;
    int magic;
} dev_request_t;

typedef struct dev_response_type
{
    unsigned int size; 
    unsigned int id; 
    unsigned int result;
} dev_response_t;


static unsigned int 
get_next_packet_id(void)
{
    static unsigned int packet_id = 0;
    if (!packet_id) {
	srand(time(NULL)^getpid());
	packet_id = rand();
    } else
	++packet_id;

    return packet_id;
}

static const char* 
get_device_file_name(sct_ctrl_t ctrl)
{
    char* name;

    assert(ctrl);

    name =  (char*)get_user_date(ctrl);

    /* sane default, for now. change it later */
    return (name ? name : "/dev/sct_ctrl");
}

static const char* 
resp_result_to_string(unsigned int res)
{
    switch (res){
    case CTL_TRACKER_SUCCESS:          return "the operation was succesfull";
    case CTL_TRACKER_EINVAL:           return "the request packet is invalid";
    case CTL_TRACKER_ENOMEM:           return "the kernel could not allocate memory";
    case CTL_TRACKER_EINTR:	       return "uperforming the request was interrupted - "
					      "the client should try again";
    case CTL_TRACKER_ENOSYS:	       return "no such syscall (invalid syscall_id parameter)";
    case CTL_TRACKER_ESYSUNSUP:        return "this syscall is not supported at the moment";
    case CTL_TRACKER_ENOENT:	       return "no such tracking rule";
    case CTL_TRACKER_EBADTRACKINGRULE: return "the tracking rule could not be deserialized";
    default:                           return "unknown response";
    }
}

static void 
free_request(dev_request_t* req)
{
    /* noop */
}

static int 
dev_file_open(const char* dev_file)
{
    assert(dev_file);

    return open(dev_file, O_RDWR);
}

static int 
dev_file_close(int fd)
{
    return close(fd);
}

static ssize_t
dev_file_read(int fd, void* buf, size_t sz)
{
    int bytes_read = 0;
    int rc;
    while (bytes_read < sz){
	if ((rc = read (fd, (unsigned char*)buf+bytes_read, 
			sz-bytes_read)) == -1){
	    if ((errno == EINTR) || (errno == EAGAIN))
		continue;
	    break;
	} else if (rc == 0) /* eof */
	    break;
	bytes_read += rc;
    }
    return (rc == -1 ? -1 : bytes_read);
}

static ssize_t
dev_file_write(int fd, void* buf, size_t sz)
{
    int bytes_written = 0;
    int rc;
    while (bytes_written < sz){
	if ((rc = write (fd, (unsigned char*)buf+bytes_written, 
			 sz-bytes_written)) == -1){
	    if ((errno == EINTR) || (errno == EAGAIN))
		continue;
	    break;
	}
	bytes_written += rc;
    }
    return (rc == -1 ? -1 : bytes_written);
}

static int 
build_request(dev_request_t* req, unsigned int cmd,
	      int rule_id, rule_when_t when, int syscall_id, 
	      unsigned char* buf, unsigned int length, int enable_rule)
{
    assert(req);

    memset(req, 0, sizeof(*req));
    req->size = sizeof(req->id) + sizeof(req->type) + sizeof(req->rule_id) +
	sizeof(req->when) + sizeof(req->syscall_id) +
	sizeof(req->tracking_rule_len) + 
	req->tracking_rule_len + sizeof(req->magic);

    req->id = get_next_packet_id();
    req->type = cmd;
    req->rule_id = rule_id;
    req->when = when;
    req->syscall_id = syscall_id;
    req->tracking_rule_len = length;
    req->tracking_rule = buf;
    req->magic = DEV_FILE_MAGIC_COOKIE;

    return 1;
}

static int 
send_request(int fd, dev_request_t* req)
{
    unsigned char buf[1000];
    int pos = 0; 

    assert(req && sizeof(buf) > req->size);

    /* no checking required, since these are all ints and we
       are adding them a 1000 bytes buffer */
    memcpy(buf, &req->size, sizeof(req->size));
    pos += sizeof(req->size);

    memcpy(buf + pos, &req->id, sizeof(req->id));
    pos += sizeof(req->id);

    memcpy(buf + pos, &req->type, sizeof(req->type));
    pos += sizeof(req->type);

    memcpy(buf + pos, &req->rule_id, sizeof(req->rule_id));
    pos += sizeof(req->rule_id);

    memcpy(buf + pos, &req->when, sizeof(req->when));
    pos += sizeof(req->when);

    memcpy(buf + pos, &req->syscall_id, sizeof(req->syscall_id));
    pos += sizeof(req->syscall_id);

    memcpy(buf + pos, &req->tracking_rule_len, sizeof(req->tracking_rule_len));
    pos += sizeof(req->tracking_rule_len);

    /* checking required, this one is variable size */
    if (req->tracking_rule_len < sizeof(buf) - pos) {
	memcpy(buf + pos, req->tracking_rule, req->tracking_rule_len);
	pos += req->tracking_rule_len;
    } else 
	return 0;

    if (sizeof(req->magic) < sizeof(buf) - pos) {
	memcpy(buf + pos, &req->magic, sizeof(req->magic));
	pos += sizeof(req->magic);
    }
    
    return (dev_file_write(fd, buf, pos) == pos);
}

static int
recv_response(int fd, dev_response_t* resp)
{
    unsigned char buf[sizeof(*resp) * 2]; /* make some room for padding? */
    int packetsz, rc;

    assert(fd >= 0 && resp);

    rc = dev_file_read(fd, (void*)buf, sizeof(*resp) * 2);
    if (rc < sizeof(resp->size))
	return 0;

    packetsz = rc;

    memcpy(&resp->size, buf, sizeof(resp->size));

    /* check that the reported packet size is equal to the
       ammount of bytes we actually read from the kernel */
    if (resp->size != packetsz)
	return 0;
    
    /* ok, we appear to have gotten a valid packet, return it */
    memcpy(resp, buf, packetsz);

    return 1;
}

static int 
parse_response(dev_response_t* resp, unsigned int request_id)
{
    assert (resp);

    /* check that it's a response for this specific request */
    if (resp->id != request_id) {
	fprintf(stderr, "'%s': got response '%d' for request '%d'\n", 
		__FUNCTION__, resp->id, request_id);
	return -1;
    }

    /* it's valid. yippee. */
    return resp->result;
}

/* we do most of the work here. we return 0 on success, 
   -1 on genereal error and >0 on one of the protocol errors */
int 
dev_file_do_protocol(sct_ctrl_t ctrl, unsigned int cmd, 
		     int rule_id, rule_when_t when, int syscall_id, 
		     unsigned char* buf, unsigned int length, 
		     int enable_rule)
{
    dev_request_t req;
    dev_response_t resp;
    int rc = -1;
    
    /* open the device file */
    int fd = dev_file_open(get_device_file_name(ctrl));
    if (fd == -1) {
	fprintf(stderr, "'%s': failed to open device file '%s' with error '%s'\n", 
		__FUNCTION__, get_device_file_name(ctrl), 
		strerror(errno));
	return rc;
    }

    if (!build_request(&req, cmd, rule_id, when, syscall_id, 
		       buf, length, enable_rule)) {
	fprintf(stderr, "'%s': failed to build a request\n", __FUNCTION__);
	goto close_device;
    }

    if (!send_request(fd, &req)){
	fprintf(stderr, "'%s': failed to send the request, with error '%s'\n", 
		__FUNCTION__, strerror(errno));
	goto free_request;
    }

    /* good, request sent, now let's wait for a response */
    if (!recv_response(fd, &resp)){
	fprintf(stderr, "'%s': failed to receive a response from the kernel, "
		"with error '%s'\n", __FUNCTION__, strerror(errno));
	goto free_request;
    }

    /* woohoo, we received a response! now, let's parse it and see what it says */
    if ((rc = parse_response(&resp, req.id)) != CTL_TRACKER_SUCCESS) {
	fprintf(stderr, "'%s': kernel returned error response: '%d - %s'\n", 
		__FUNCTION__, resp.result, resp_result_to_string(resp.result));
	if (rc != -1)
	    rc = resp.result;
    }

    /* cleanup and return success or whatever error we got */
 free_request:
    free_request(&req);

 close_device:
    dev_file_close(fd);
    
    return rc;
}
