/*
 *$Id: sct_ctrl_lib.c,v 1.13 2001/12/09 19:55:06 chooga Exp $
 */
#include "sct_ctrl_lib.h"
#include "sct_ctrl_lib_private.h"
#include "sysctl_impl.h"
#include "dev_file_impl.h"

/* get the module's sysctl defines */
#include "sct_rules_module/module_interface.h"

#include <malloc.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

/* init the library */
sct_ctrl_t
sct_ctrl_init(void)
{
    sct_ctrl_t ctrl = malloc(sizeof(*ctrl));
    if (!ctrl){
	return NULL;
    }

    /* better safe than sorry */
    memset(ctrl, 0, sizeof(struct sct_ctrl_type));

    /* default implementation type */
    ctrl->impl = SCT_CTRL_SYSCTL;

    return ctrl;
}

/* shutdown the library */
int sct_ctrl_cleanup(sct_ctrl_t ctrl)
{
    if (ctrl)
	free(ctrl);

    return 1;
}

/* add a rule with the given rule id to the given syscall,    */
/* where the rule is made of this filter and this action      */
/* and 'enable rule' specifies if the rule should be          */
/* immediately enabled or not                                 */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_add_rule(sct_ctrl_t ctrl, int rule_id, rule_when_t when,
		  int syscall_id, 
		  filter_node_t* filter, action_t* act, 
		  int enable_rule)
{
    if (!ctrl)
	return 0;

    switch (ctrl->impl){
    case (SCT_CTRL_SYSCTL):
	return sysctl_add_rule(ctrl, rule_id, when, syscall_id, filter, 
			       act, enable_rule);
    case (SCT_CTRL_DEVICE_FILE):
	return dev_file_add_rule(ctrl, rule_id, when, syscall_id, filter, 
				 act, enable_rule);
    default:
	fprintf(stderr, "%s: unknown control library implementation '%d'\n",
		__FUNCTION__, ctrl->impl);
	return 0;
    }
}

/* delete from the given syscall the rule with this rule id   */
/* return 0 on failure, non 0 on success                      */
int
sct_ctrl_del_rule(sct_ctrl_t ctrl,
		  int rule_id,
		  rule_when_t when,
		  int syscall_id)
{
    if (!ctrl)
	return 0;

    switch (ctrl->impl){
    case (SCT_CTRL_SYSCTL):
	return sysctl_del_rule(ctrl, rule_id, when, syscall_id);
	
    case (SCT_CTRL_DEVICE_FILE):
	return dev_file_del_rule(ctrl, rule_id, when, syscall_id);

    default:
	fprintf(stderr, "%s: unknown control library implementation '%d'\n",
		__FUNCTION__, ctrl->impl);
	return 0;
    }
}

/* enable the rule with this id for this syscall              */
/* NOTE: currently NOT IMPLEMENTED                            */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_enable_rule(sct_ctrl_t ctrl,
		     int rule_id,
		     rule_when_t when,
		     int syscall_id)
{
    if (!ctrl)
	return 0;

    switch (ctrl->impl){
    case (SCT_CTRL_SYSCTL):
	return sysctl_enable_rule(ctrl, rule_id, when, syscall_id);
	
    case (SCT_CTRL_DEVICE_FILE):
	return dev_file_enable_rule(ctrl, rule_id, when, syscall_id);

    default:
	fprintf(stderr, "%s: unknown control library implementation '%d'\n",
		__FUNCTION__, ctrl->impl);
	return 0;
    }
}

/* disable the rule with this id for this syscall             */
/* NOTE: currently NOT IMPLEMENTED                            */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_disable_rule(sct_ctrl_t ctrl,
		      int rule_id,
		      rule_when_t when,
		      int syscall_id)
{
    if (!ctrl)
	return 0;

    switch (ctrl->impl){
    case (SCT_CTRL_SYSCTL):
	return sysctl_disable_rule(ctrl, rule_id, when, syscall_id);
	
    case (SCT_CTRL_DEVICE_FILE):
	return dev_file_disable_rule(ctrl, rule_id, when, syscall_id);

    default:
	fprintf(stderr, "%s: unknown control library implementation '%d'\n",
		__FUNCTION__, ctrl->impl);
	return 0;
    }
}

/* delete all of the rules for this system call.              */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_del_all_syscall_rules(sct_ctrl_t ctrl, int syscall_id)
{
    if (!ctrl)
	return 0;

    switch (ctrl->impl){
    case (SCT_CTRL_SYSCTL):
	return sysctl_del_all_syscall_rules(ctrl, syscall_id);
	
    case (SCT_CTRL_DEVICE_FILE):
	return dev_file_del_all_syscall_rules(ctrl, syscall_id);
	
    default:
	fprintf(stderr, "%s: unknown control library implementation '%d'\n",
		__FUNCTION__, ctrl->impl);
	return 0;
    }
}

/* delete all rules for ALL syscalls.                         */
/* return 0 on failure, non 0 on success                      */
int 
sct_ctrl_del_all_rules(sct_ctrl_t ctrl)
{
    if (!ctrl)
	return 0;

    switch (ctrl->impl){
    case (SCT_CTRL_SYSCTL):
	return sysctl_del_all_rules(ctrl);
	
    case (SCT_CTRL_DEVICE_FILE):
	return dev_file_del_all_rules(ctrl);
	
    default:
	fprintf(stderr, "%s: unknown control library implementation '%d'\n",
		__FUNCTION__, ctrl->impl);
	return 0;
    }
}

/* print all of the current rules to the system's log file    */
/* return 0 on failure, non 0 on success                      */
int
sct_ctrl_print_all_rules(sct_ctrl_t ctrl)
{
    if (!ctrl)
	return 0;

    switch (ctrl->impl){
    case (SCT_CTRL_SYSCTL):
	return sysctl_print_all_rules(ctrl);
	
    case (SCT_CTRL_DEVICE_FILE):
	return dev_file_print_all_rules(ctrl);
	
    default:
	fprintf(stderr, "%s: unknown control library implementation '%d'\n",
		__FUNCTION__, ctrl->impl);
	return 0;
    }
}






