/*
 *$Id: filter_vars.c,v 1.7 2001/12/08 09:24:39 amirs Exp $
 */

#include "mem_manager_public.h"
#include "filter_vars_public.h"

/* constants. */
#define MAX_PARAM_NUM 5		/* highest supported parameter. */

/* var struct - contains data about each var */
struct var_struct {
    int persistent;
    void *value;
    void *data;
    data_type_t dtype;
    union {
	callback_func var;
	callback_indexed_func indexed;
    } f;
};

/* list of variables data. */
struct var_struct vars[VT_LAST_TYPE];
struct var_struct params[MAX_PARAM_NUM];
static int var_initialized = 0;

/*
 * Private functions.
 */

/* get the value of the given parameter */
filter_node_t*
get_param(int param_index)
{
    filter_node_t* val = NULL;

    if (var_initialized && (param_index >= 0 && param_index < MAX_PARAM_NUM)) {
	val = params[param_index].value;
	if (val == NULL && params[param_index].f.indexed != NULL) {
	    val = params[param_index].f.indexed(VT_PARAM, param_index, params[param_index].dtype, params[param_index].data);
	    params[param_index].value = val;
	    /* persistent callback support for params isn't needed for now */
	    params[param_index].f.indexed = NULL; 
	}
    }

    return val;
}

/* set the value of the given parameter to the given value. */
void
set_param(int param_index, filter_node_t* param_val)
{
    /* range checking. */
    if (param_index < 0 || param_index >= MAX_PARAM_NUM)
	return;

    free_filter_node(params[param_index].value);
    params[param_index].value = param_val;
}

/* set the value of the given parameter to the callback function */
void
set_param_callback(int param_index, callback_indexed_func func, data_type_t dtype, void *data)
{
    /* range checking. */
    if (param_index < 0 || param_index >= MAX_PARAM_NUM)
	return;

    params[param_index].f.indexed = func;
    params[param_index].data = data;
    params[param_index].dtype = dtype;
}

/* free all allocated parameter values. */
void
free_params_list(void)
{
    int i;
    for (i=0; i<MAX_PARAM_NUM; i++) {
        free_filter_node(params[i].value);
	params[i].value = NULL;
	params[i].f.indexed = NULL;
	params[i].data = NULL;
	params[i].dtype = DT_LAST_TYPE;
    }
}

/* initialized the parameter array. */
void
init_params(void)
{
    int i;
    for (i=0; i<MAX_PARAM_NUM; i++) {
	params[i].value = NULL;
	params[i].f.indexed = NULL;
	params[i].data = NULL;
	params[i].dtype = DT_LAST_TYPE;
	params[i].persistent = 0;
    }
}

/* initialized the variables array. */
void
init_vars(void)
{
    int i;
    
    init_params();
    for (i=0; i<VT_LAST_TYPE; i++) {
	vars[i].value = NULL;
	vars[i].f.var = NULL;
	vars[i].data = NULL;
	vars[i].dtype = DT_LAST_TYPE;
	vars[i].persistent = 0;
    }
    
    var_initialized = 1;
}


/*
 * Kernal variables manipulation code.
 */

/* empty variable values array. */
void
reset_all_vars(void)
{
    int i;

    if (!var_initialized) {
	init_vars();
    }

    for (i=0; i<VT_LAST_TYPE; i++) {
	switch (i) {
	case VT_PARAM:
	    free_params_list();
	    break;
	default:
	    free_filter_node((filter_node_t*)vars[i].value);
	    vars[i].value = NULL;
	    if (vars[i].persistent == 0) {
	        vars[i].f.var = NULL;
		vars[i].data = NULL;
		vars[i].dtype = DT_LAST_TYPE;
	    }
	    break;
	}
    }
}

/* store a given variable's value. */
void set_var(var_type_t type, filter_node_t* val)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    /* this is an indexed variable. */
	    break;
	default:
	    vars[type].value = val;
	    break;
	}
    }
}

/* store a given variable's callback function. */
static void _set_var_callback(var_type_t type, int persistent, callback_func func, void* data)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    /* this is an indexed variable. */
	    break;
	default:
	    vars[type].f.var = func;
	    vars[type].data = data;
	    vars[type].persistent = persistent;
	    break;
	}
    }
}

void set_var_callback(var_type_t type, callback_func func, void* data)
{
    return _set_var_callback(type, 0, func, data);
}

void set_persistent_callback(var_type_t type, callback_func func, void* data)
{
    return _set_var_callback(type, 1, func, data);
}

void set_var_indexed(var_type_t type, int var_index, filter_node_t* val)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    set_param(var_index, val);
	    break;
	default:
	    /* these are not indexed variables. */
	    break;
	}
    }
}

void set_var_indexed_callback(var_type_t type, int var_index, callback_indexed_func func, data_type_t dtype, void *data)
{
    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch (type) {
	case VT_PARAM:
	    set_param_callback(var_index, func, dtype, data);
	    break;
	default:
	    /* these are not indexed variables. */
	    break;
	}
    }
}

/* evaluate the value of the given parameter, assuming it was previously */
/* set with set_var/set_var_index.                                       */
filter_node_t* get_var(var_type_t type)
{
    filter_node_t* val = NULL;

    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch(type) {
	case VT_PARAM:
	    /* VT_PARAM is an indexed var. */
	    break;
	default:
	    val = vars[type].value;
	    if (val == NULL && vars[type].f.var != NULL) {
		val = vars[type].f.var(type, vars[type].data);
		vars[type].value = val;
		if (vars[type].persistent == 0) 
		    vars[type].f.var = NULL;
	    }
	    break;
	}
    }

    return val;
}

filter_node_t* get_var_indexed(var_type_t type, int var_index)
{
    filter_node_t* val = NULL;

    if (type > 0 && type < VT_LAST_TYPE && var_initialized) {
	switch(type) {
	case VT_PARAM:
	    val = get_param(var_index);
	    break;
	default:
	    /* these are not indexed vars */
	    break;
	}
    }

    return val;
}

