/*
 *$Id: rule_engine.c,v 1.25 2001/12/19 07:57:21 amirs Exp $
 */
#include "mem_manager_public.h"
#include "encode_decode_public.h"
#include "rule_engine_private.h"
#include "filter_vars_public.h"
#include "archdep.h"

#include "stdarg.h"

/* a list of tracking rules. */
#define	MAX_NUM_TRACKER_RULES	20

struct rule_list {
    struct tracker_rule* rules[MAX_NUM_TRACKER_RULES];
    int num_rules;
};

/* maximum size of a log record. */
#define MAX_LOG_RECORD	1024

/* types of logging macros, used when logging syscall invocations. */
typedef enum {
    LM_PID = 0,
    LM_UID = 0,
    LM_EUID = 0,
} log_macro_t;

/*
 * Private functions.
 */

/* find the location of a rule with the given ID in the rules list. */
/* returns the rule's location, or '-1' if none is found.           */
int
find_tracking_rule(rule_list_t* rules, int rule_id)
{
    int i;

    if (!rules)
	return -1;

    for (i = 0; i < rules->num_rules; i++) {
	if (rules->rules[i] && rules->rules[i]->rule_id == rule_id)
	    break;
    }

    return (i < rules->num_rules ? i : -1);
}

/* return a copy of the 'src' action */
void copy_action(action_t* dst, const action_t* src)
{
    action_t* tmp = NULL;; 
    if (src) {
	switch (src->type) {
	case AT_LOG: 
	    tmp = make_action_log(src->u.log.priority, src->u.log.format);
	    break;
	case AT_RULE_ENABLE: 
	    tmp = make_action_enable_rule(src->u.rule_id);
	    break;
	case AT_RULE_DISABLE: 
	    tmp = make_action_disable_rule(src->u.rule_id);
	    break;
	case AT_PROCESS_SUSPEND: 
	    tmp = make_action_suspend_process();
	    break;
	case AT_PROCESS_KILL: 
	    tmp = make_action_kill_process();
	    break;
	case AT_SYSCALL_FAIL: 
	    tmp = make_action_fail_syscall(src->u.error_code);
	    break;
	default:
	    PPRINT("copy_action(): unknown action type: %d\n", src->type);
	    return;
	}
	/* ugh, not nice */
	memcpy(dst, tmp, sizeof(*dst));
	/* _not_ action_free() */
	mem_free(tmp);
    }
}

/* add the given rule to the rules list.                                */
/* returns the rule's location on success, or '-1' if an error occured. */
int
do_add_tracking_rule(rule_list_t* rules,
		     int rule_id,
		     filter_node_t* filter,
		     action_t* action)
{
    int i;
    int j;
    struct tracker_rule* rule;

    /* make sure we have free space. */
    if (!rules || rules->num_rules >= MAX_NUM_TRACKER_RULES) {
	return -1;
    }

    /* allocate a structure for the new rule. */
    rule = mem_alloc(sizeof(struct tracker_rule));
    if (!rule) {
	return -1;
    }

    rule->rule_id = rule_id;
    rule->filter = filter;
    memset(&rule->action, 0, sizeof(rule->action));
    copy_action(&rule->action, action);
	
    /* find the right location for the new rule. */
    for (i = 0; i < rules->num_rules; i++) {
	if (rules->rules[i] && rules->rules[i]->rule_id > rule_id)
	    break;
    }
	
    /* make a free space for the new rule. */
    for (j = rules->num_rules; j > i; j--) {
	rules->rules[j] = rules->rules[j-1];
    }
	
    /* insert the new rule. */
    rules->rules[i] = rule;
    rules->num_rules++;
	
    return (i);
}

/* serialize a 'action_t' action node. */
int
do_serialize_action_node(action_t* a,
			 unsigned char* buf,
			 int buflen,
			 unsigned char** pos)
{
    int rc = 0;

    if (a) {
	char action_type = a->type;
	rc = encode_char(buf, buflen, pos, action_type);
	if (rc > 0) {
	    switch (action_type) {
	    case AT_LOG:
		rc = encode_int(buf,
				buflen,
				pos,
				a->u.log.priority);
		if (rc > 0) {
		    rc = encode_string(buf,
				       buflen,
				       pos,
				       a->u.log.format);
		}
		break;
	    case AT_RULE_ENABLE:
		rc = encode_int(buf,
				buflen, pos, a->u.rule_id);
		break;
	    case AT_RULE_DISABLE:
		rc = encode_int(buf,
				buflen, pos, a->u.rule_id);
		break;
	    case AT_PROCESS_SUSPEND:
		/* no data for this action. */
		break;
	    case AT_PROCESS_KILL:
		/* no data for this action. */
		break;
	    case AT_SYSCALL_FAIL:
		rc = encode_int(buf,
				buflen, pos, a->u.error_code);
		break;
	    default:
		rc = 0;
		break;
	    }
	}
    }
    else {
	rc = encode_char(buf, buflen, pos, AT_NIL);
    }

    return rc;
}

/* de-serialize an 'action_t' action node. */
int
do_deserialize_action_node(action_t** pa,
			   unsigned char* buf,
			   int buflen,
			   unsigned char** pos)
{
    char action_type;
    int rc = decode_char(buf, buflen, pos, &action_type);

    if (rc > 0) {
	switch (action_type) {
	case AT_LOG:
	{
	    int priority;
	    char* format;
	    rc = decode_int(buf,
			    buflen, pos, &priority);
	    if (rc > 0) {
		rc = decode_string(buf,
				   buflen,
				   pos,
				   &format);
	    }
	    if (rc > 0) {
		*pa = make_action_log(priority,
				      format);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_RULE_ENABLE:
	{
	    int rule_id;
	    rc = decode_int(buf,
			    buflen, pos, &rule_id);
	    if (rc > 0) {
		*pa = make_action_enable_rule(
		    rule_id);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_RULE_DISABLE:
	{
	    int rule_id;
	    rc = decode_int(buf,
			    buflen, pos, &rule_id);
	    if (rc > 0) {
		*pa = make_action_disable_rule(
		    rule_id);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_PROCESS_SUSPEND:
	{
	    *pa = make_action_suspend_process();
	    if (!*pa) {
		rc = 0;
	    }
	}
	break;
	case AT_PROCESS_KILL:
	{
	    *pa = make_action_kill_process();
	    if (!*pa) {
		rc = 0;
	    }
	}
	break;
	case AT_SYSCALL_FAIL:
	{
	    int error_code;
	    rc = decode_int(buf,
			    buflen,
			    pos,
			    &error_code);
	    if (rc > 0) {
		*pa = make_action_fail_syscall(
		    error_code);
		if (!*pa) {
		    rc = 0;
		}
	    }
	}
	break;
	case AT_NIL:
	    /* a NULL action_node. */
	    *pa = NULL;
	    rc = 1;
	    break;
	default:
	    /* shouldn't happen. */
	    rc = 0;
	    break;
	}
    }

    return rc;
}


/*
 * Functions for actions manipulation.
 */

/* allocate an action object for logging purposes. */
action_t*
make_action_log(int priority, const char* format)
{
/* OPTIMIZATION: keep static actions and return pointers to them */

    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->type = AT_LOG;
	action->u.log.priority = priority;
	if (format) {
	    action->u.log.format =
		(char*)mem_alloc(strlen(format)+1);
	    if (action->u.log.format)
		strcpy(action->u.log.format, format);
	    else {
		mem_free(action);
		action = NULL;
	    }
	}
	else {
	    action->u.log.format = NULL;
	}
    }

    return action;
}

/* allocate an action object for rule enabling purposes. */
action_t*
make_action_enable_rule(int rule_id)
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->type = AT_RULE_ENABLE;
	action->u.rule_id = rule_id;
    }

    return action;
}

/* allocate an action object for rule disabling purposes. */
action_t*
make_action_disable_rule(int rule_id)
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->type = AT_RULE_DISABLE;
	action->u.rule_id = rule_id;
    }

    return action;
}

/* allocate an action object for process suspending. */
action_t* make_action_suspend_process()
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->type = AT_PROCESS_SUSPEND;
    }

    return action;
}

/* allocate an action object for process killing. */
action_t* make_action_kill_process()
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->type = AT_PROCESS_KILL;
    }

    return action;
}

/* allocate an action object for system call failing. */
action_t* make_action_fail_syscall(int error_code)
{
    struct tracker_action* action =
	mem_alloc(sizeof(struct tracker_action));
    if (action) {
	action->type = AT_SYSCALL_FAIL;
	action->u.error_code = error_code;
    }

    return action;
}

void free_action(action_t* a)
{
    if (a) {
	switch(a->type) {
	case AT_LOG:
	    if (a->u.log.format)
		mem_free(a->u.log.format);
	    break;
	default: break; /* do nothing */
	}
	mem_free(a);
    }
}


/*
 * Action serialization/de-serialization functions.
 */

/* serialize an action node into a buffer, for export     */
/* purposes. returns the size of the serialized data, 0   */
/* if there was an error, -1 if buffer is too small.      */
int
serialize_action_node(action_t* a, unsigned char* buf, int buflen)
{
    unsigned char* pos = buf;
    int rc = do_serialize_action_node(a, buf, buflen, &pos);

    return (rc > 0 ? pos - buf : rc);
}

/* de-serialize an action node from a buffer, for import  */
/* purposes. returns the new allocated action node or     */
/* NULL if an error occuerd.                              */
action_t*
deserialize_action_node(unsigned char* buf, int buflen)
{
    unsigned char* pos = buf;
    action_t* a = NULL;
    int rc = 0;

    if (buflen > 0) {
	rc = do_deserialize_action_node(&a, buf, buflen, &pos);
    }

    return (rc > 0 ? a : NULL);
}

/* print an action node to a "standard" device */
void print_action_node(action_t* a)
{
    char* action_name;

    switch (a->type) {
    case AT_LOG:
	action_name = "log";
	break;
    case AT_RULE_ENABLE:
	action_name = "enable rule";
	break;
    case AT_RULE_DISABLE:
	action_name = "disable rule";
	break;
    case AT_PROCESS_SUSPEND:
	action_name = "suspend process";
	break;
    case AT_PROCESS_KILL:
	action_name = "kill process";
	break;
    case AT_SYSCALL_FAIL:
	action_name = "fail syscall";
	break;
    default:
	action_name = "(unknown)";
	break;
    }
    PPRINT(" Action: %s\n", action_name);
}


/*
 * Functions for tracking rules manipulation.
 */

/* serialize a tracking rule into a buffer, for export   */
/* purposes. returns the size of the serialized data, 0  */
/* if there was an error, -1 if the buffer is too small. */
int serialize_tracking_rule(int rule_id, rule_when_t when, int syscall_id,
			    action_t* a, filter_node_t* f,
		            unsigned char* buf, int buflen)
{
    unsigned char* pos_1 = buf;
    int rc = encode_int(buf, buflen, &pos_1, rule_id);
    if (rc > 0) {
	rc = encode_int(buf, buflen, &pos_1, when);
        if (rc > 0) {
	    rc = encode_int(buf, buflen, &pos_1, syscall_id);
	    if (rc > 0) {
	        int action_size =
		    serialize_action_node(a, pos_1, buflen-(pos_1-buf));
	        if (action_size > 0) {
		    unsigned char* pos_2 = pos_1+action_size;
		    int filter_size =
		        serialize_filter_node(f,
					      pos_2,
					      buflen-(pos_2-buf));
		    if (filter_size > 0) {
		        rc = (pos_2 - buf) + filter_size;
		    }
		    else {
		        rc = filter_size;
		    }
	        }
	        else {
		    rc = action_size;
	        }
	    }
        }
    }

    return rc;
}

/* de-serialize a tracking rule from a buffer, for import */
/* purposes. returns the 1 for success, -1 for failure.   */
/* the given action rule and filter node are filled with  */
/* elements geenrated from the buffer, and should be      */
/* freed by the caller.                                   */
int deserialize_tracking_rule(unsigned char* buf, int buflen, 
			      int* p_rule_id, rule_when_t* p_when,
			      int* p_syscall_id,
			      action_t** pa, filter_node_t** pf)
{
    unsigned char* pos_1 = buf;
    int rc = decode_int(buf, buflen, &pos_1, p_rule_id);

    if (rc > 0) {
	rc = decode_int(buf, buflen, &pos_1, (int*)p_when);
        if (rc > 0) {
	    rc = decode_int(buf, buflen, &pos_1, p_syscall_id);
	    if (rc > 0) {
	        unsigned char* pos_2 = pos_1;
	        rc = do_deserialize_action_node(pa,
					        pos_1,
					        buflen-(pos_1-buf),
					        &pos_2);
	        if (rc > 0) {
		    unsigned char* pos_3 = pos_2;
		    rc = do_deserialize_filter_node(pf,
						    pos_2,
						    buflen-(pos_2-buf),
						    &pos_3);
		    if (rc == 0) { 
			free_action(*pa);
		    }
	        }
	    }
        }
    }

    return rc;
}

/* create an empty tracking rules list. */
rule_list_t*
create_tracking_rule_list()
{
    rule_list_t* rules = mem_alloc(sizeof(struct rule_list));
    if (rules) {
	rules->num_rules = 0;
    }

    return rules;
}

/* delete the given tracking rule list, freeing any rule found in it. */
void
delete_tracking_rule_list(rule_list_t* rules)
{
    del_all_tracking_rules_in_list(rules);
    mem_free(rules);
}

/* add a new tracking rule. returns the new rule's id */
/* returns '0' if an error occured, '-1' if there is  */
/* already a rule with the given rule ID.             */
int
add_tracking_rule(rule_list_t* rules,
		  int rule_id,
		  filter_node_t* filter,
		  action_t* action)
{
    /* assume failure. */
    int res = 0;

    if (filter && action) {
	/* make sure we don't already have a rule with the same ID. */ 
	int rule_loc = find_tracking_rule(rules, rule_id);
	if (rule_loc >= 0) {
	    res = -1;
	}
	else {
	    /* add the rule, sorted by rule ID. */
	    res = do_add_tracking_rule(rules,
				       rule_id,
				       filter,
				       action);
	}
    }

    return res;
}

/* delete a tracking rule. */
void
del_tracking_rule(rule_list_t* rules, int rule_id)
{
    int i = find_tracking_rule(rules, rule_id);

    if (i >= 0 && i < rules->num_rules) {
	struct tracker_rule* rule = rules->rules[i];
	int j;

	/* we're removing a rule - close the 'gap' created in the list. */
	for (j = i; j < rules->num_rules-1; j++) {
	    rules->rules[j] = rules->rules[j+1];
	}
	if (rule->filter) {
	    free_filter_node(rule->filter);
	}
	mem_free(rule);
	rules->num_rules--;
    }
}

/* deletes all tracking rules in the list. */
void
del_all_tracking_rules_in_list(rule_list_t* rules)
{
    int i;

    for (i = 0; i < rules->num_rules; i++) {
	if (rules->rules[i]) {
	    if (rules->rules[i]->filter) {
	        free_filter_node(rules->rules[i]->filter);
	    }
	    mem_free(rules->rules[i]);
	    rules->rules[i] = NULL;
	}
    }
    rules->num_rules = 0;
}

/* check if there is a rule with the given ID. */
/* returns '1' if there is, '0' otehrwise.     */
int
has_tracking_rule(rule_list_t* rules, int rule_id)
{
    int i = find_tracking_rule(rules, rule_id);

    return (i >= 0 ? 1 : 0);
}

/* returns the number of rules in the given rule list. */
int
get_tracking_rule_count(rule_list_t* rules)
{
    return (rules ? rules->num_rules : 0);
}

/* print out the list of current tracking rules. */
void print_tracking_rules(rule_list_t* rules)
{
    int i;

    PPRINT("\nActive Tracking Rules List (%d):\n", rules->num_rules);
    for (i=0; i < rules->num_rules; i++) {
	/* skip empty rules */
	if (!rules->rules[i])
	    continue;
	PPRINT(" Rule id: %d\n", rules->rules[i]->rule_id);
	print_action_node(&rules->rules[i]->action);
	PPRINT(" Filter:\n");
	print_filter_node(rules->rules[i]->filter);
    }
    PPRINT("\n");
}

/*
 * Functions used by tracking functions to perform filtering.
 */

/* filter an incoming system call.                                          */
/* accepts a system call ID and its (var_args) parameters.                  */
/* returns a tracker_action structure if there was a match, NULL otherwise. */
/* on successfull match, the rule id is set into '*rule_id'.                */
action_t* filter_syscall(rule_list_t* rules, int syscall_id, int* rule_id, ...)
{
    int i;
    action_t* action = NULL;
    filter_node_t* result = make_bool_val(TF_FALSE);

    for (i=0; i<rules->num_rules; i++) {
	if (rules->rules[i] && rules->rules[i]->filter) {
	    exp_error_t rc =
		eval_filter_node(rules->rules[i]->filter, result);
	    if (rc == ERR_NONE && result->data_type == DT_BOOL &&
		result->u.constant.u.b)
		break;
	}
    }

    if (i < rules->num_rules) {
	action = &rules->rules[i]->action;
	if (rule_id)
	    *rule_id = rules->rules[i]->rule_id;
    }

    free_filter_node(result);

    return action;
}

/* stringify the given unsigned numeric value, into the given buffer, */
/* with the given size. returns 1 on succes, 0 on buffer overflow.     */
static int
tracker_stringify_unum(unsigned long num, char* buf, int buflen)
{
    char sbuf[MAX_LOG_RECORD+1];
    int rc = 0;
    
    if (buflen > 32) {
        rc = sprintf(sbuf, "%lu", num);
        if (rc < buflen) {
	    strcpy(buf, sbuf);
        }
    }

    return (rc > 0 && rc < buflen ? rc : 0);
}

/* stringify the given pointer value, into the given buffer,      */
/* with the given size. rturns 1 on succes, 0 on buffer overflow. */
static int
tracker_stringify_pointer(void* p, char* buf, int buflen)
{
    char sbuf[MAX_LOG_RECORD+1];
    int rc = 0;
    
    if (buflen > 32) {
        rc = sprintf(sbuf, "%p", p);
        if (rc < buflen) {
	    strcpy(buf, sbuf);
        }
    }

    return (rc > 0 && rc < buflen ? rc : 0);
}

/* stringify the given numeric value, into the given buffer, with the */
/* given size. returns length of stringified num on succes, 0 on      */
/* buffer overflow.                                                   */
static int
tracker_stringify_num(long num, char* buf, int buflen)
{
    char sbuf[MAX_LOG_RECORD+1];
    int rc = 0;
    
    if (buflen > 32) {
        rc = sprintf(sbuf, "%ld", num);
        if (rc < buflen) {
	    strcpy(buf, sbuf);
        }
    }

    return (rc > 0 && rc < buflen ? rc : 0);
}

/* stringify the given variable into the given buffer.                */
/* returns the length of the stringified var, '0' on buffer overflow. */
static int
tracker_stringify_var_node(char* buf, int buflen, filter_node_t *var_val)
{
    int rc = 0;
    int len = 0;
    char* string = NULL;
    char* pb;

    if (var_val && var_val->type == NT_CONST) {
	switch (var_val->data_type) {
	case DT_USHORT:
	    rc = tracker_stringify_unum(CONST_USHORT(var_val), buf, buflen);
	    break;
	case DT_SHORT:
	    rc = tracker_stringify_num(CONST_SHORT(var_val), buf, buflen);
	    break;
	case DT_UINT:
	    rc = tracker_stringify_unum(CONST_UINT(var_val), buf, buflen);
	    break;
	case DT_INT:
	    rc = tracker_stringify_num(CONST_INT(var_val), buf, buflen);
	    break;
	case DT_ULONG:
	    rc = tracker_stringify_unum(CONST_ULONG(var_val), buf, buflen);
	    break;
	case DT_LONG:
	    rc = tracker_stringify_num(CONST_LONG(var_val), buf, buflen);
	    break;
	case DT_STRING:
	    string = CONST_STRING(var_val);
	    pb = buf;
	    if (pb - buf < buflen) {
	        *pb++ = '"';
	    }
	    if (string) {
		if (strlen(string) < buflen-(pb-buf)) {
		    strcpy(pb, string);
		    pb += strlen(string);
		}
	    }
	    else {
		if (strlen("(nil)") < buflen-(pb-buf)) {
		    strcpy(pb, "(nil)");
		    pb += strlen("(nil)");
		}
	    }
	    if (pb - buf < buflen) {
	        *pb++ = '"';
	    }
	    if (pb - buf < buflen)
		*pb = '\0';
	    else
		*(pb - 1) = '\0';
	    rc = 1;
	    break;
	default:
	    break;
	}
    }

    if (rc)
	len = strlen(buf);

    return (rc ? len : 0);
}

static int
tracker_stringify_var(char* buf, int buflen, var_type_t var)
{
    return tracker_stringify_var_node(buf, buflen, get_var(var));
}

static int
tracker_stringify_var_indexed(char* buf, int buflen, var_type_t var, int var_index)
{
    return tracker_stringify_var_node(buf, buflen, get_var_indexed(var, var_index));
}

/* stringify the parameters of the syscall. */
static int 
tracker_stringify_params(char* buf,
			 int buflen,
			 const char* param_spec,
			 va_list ap)
{
    char* pb = buf;
    char c;
    unsigned char uc;
    long l;
    unsigned long ul;
    int rc;
    void* v, *p;

    if (param_spec) {
	int next_is_unsigned = 0;
	char* string = NULL; 
	char this_char = 0; 
	char next_char = 0; 
	int is_string, is_struct; /* strings and struct require a bit of special handling */ 
	int skip_params = 0; /* are we in 'parameters skipping' mode? */

	while (*param_spec && pb - buf < buflen) {
	    this_char = *param_spec++;
	    is_string = (this_char == 's');
	    is_struct = (this_char == 'S');

	    switch (this_char){
	    case ' ': break; /* skip spaces */
	    case 'u': next_is_unsigned = 1; break;
	    case 'c':
		if (next_is_unsigned) {
		    uc = (unsigned char)va_arg(ap, unsigned int);
		    if (!skip_params) {
		        if (pb - buf < buflen)
		            *pb++ = uc;
		    }
		    next_is_unsigned = 0;
		} else {
		    c = (char)va_arg(ap, int);
		    if (!skip_params) {
		        if (pb - buf < buflen)
		            *pb++ = c;
		    }
		}
		break;
            case 'h':
		if (next_is_unsigned) {
		    ul = (unsigned short)va_arg(ap, unsigned int);
		    if (!skip_params) {
		        rc = tracker_stringify_unum(ul, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		    next_is_unsigned = 0;
		} else {
		    l = (short)va_arg(ap, int);
		    if (!skip_params) {
		        rc = tracker_stringify_num(l, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		}
		break;
	    case 'i':
		if (next_is_unsigned) {
		    ul = va_arg(ap, unsigned int);
		    if (!skip_params) {
		        rc = tracker_stringify_unum(ul, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		    next_is_unsigned = 0;
		} else {
		    l = va_arg(ap, int);
		    if (!skip_params) {
		        rc = tracker_stringify_num(l, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		}
		break;
	    case 'l':
		if (next_is_unsigned) {
		    ul = va_arg(ap, unsigned long);
		    if (!skip_params) {
		        rc = tracker_stringify_unum(ul, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		    next_is_unsigned = 0;
		} else {
		    l = va_arg(ap, long);
		    if (!skip_params) {
		        rc = tracker_stringify_num(l, pb, buflen-(pb-buf));
		        pb += rc;
		    }
		}
		break;
	    case 'p':
		p = va_arg(ap, void*);
		if (!skip_params) {
		    rc = tracker_stringify_pointer(p, pb, buflen-(pb-buf));
		    pb += rc;
		}
		break;
	    case 'P': /* indexed var */
		l = va_arg(ap, int);
		if (!skip_params) {
		    rc = tracker_stringify_var_indexed(pb, buflen-(pb-buf), VT_PARAM, l);
		    pb += rc;
		}
		break;
	    case 's': /* string */
	    case 'S': /* struct */
		string = va_arg(ap, char*);
		if (!skip_params) {
		    /* for a string, we want to output '"string"' */
		    if (is_string && (pb - buf < buflen))
		        *pb++ = '"';
		    if (string) {
		        if (strlen(string) < buflen-(pb-buf)) {
			    strcpy(pb, string);
			    pb += strlen(string);
		        }
		    }
		    else { /* the string was null */
		        if (strlen("(nil)") < buflen-(pb-buf)) {
			    strcpy(pb, "(nil)");
			    pb += strlen("(nil)");
		        }
		    }
		    /* for a string, we want to output '"string"' */
		    if (is_string) {
			if (pb - buf < buflen) {
			    *pb++ = '"';
			}
			/* for strings, we are done */
			break;
		    }
		    /* for structures, the next arg is a pointer to the _user space_ structure
		       we only care if it's NULL or not, and we never ever dereference it */
		    v = va_arg(ap, void*);
		    if (v == NULL)
			skip_params = 1;
		}
		break;
	    case '{':
		if (pb - buf < buflen)
		    *pb++ = '{';
		break;
	    case '}':
		if (pb - buf < buflen)
		    *pb++ = '}';
		if (skip_params)
		    skip_params = 0;
		break;
	    case 'v':
		if (!skip_params) {
		    if (strlen("void") < buflen-(pb-buf)) {
		        strcpy(pb, "void");
		        pb += strlen("void");
		    }
		}
		break;
	    default:
		break;
	    }
	    /* if we just printed a value, print a comma after it */
	    next_char = *param_spec;
	    if ((this_char != ' ') && (this_char != 'u') && (this_char != 'S') && (this_char != '{') && (next_char) && next_char != '}') {
		if (!skip_params) {
		    if (strlen(", ") < buflen-(pb-buf)) {
		        strcpy(pb, ", ");
		        pb += strlen(", ");
		    }
		}
	    }
	}
    }

    return pb - buf;
}

/* perform macro translation into the given buffer with the given size. */
/* returns the size of data writtern into the buffer.                   */
static int 
tracker_translate_macro(rule_list_t* rules, 
			int rule_id, 
			int syscall_id, 
			const char* syscall_name, 
			const char* macro, 
			char* buf, 
			int buflen, 
			const char* param_spec, 
			va_list ap)
{
    int rc;
    int len = 0;

    if (!strcmp(macro, "ruleid")) {
	rc = tracker_stringify_num(rule_id, buf, buflen);
	if (rc)
	    len = strlen(buf);
    }
    else if (!strcmp(macro, "sid")) {
	rc = tracker_stringify_num(syscall_id, buf, buflen);
	if (rc)
	    len = strlen(buf);
    }
    else if (!strcmp(macro, "sname")) {
	if (strlen(syscall_name) < buflen - 1) {
	    strcpy(buf, syscall_name);
	    len = strlen(buf);
	}
    }
    else if (!strcmp(macro, "params")) {
	rc = tracker_stringify_params(buf, buflen, param_spec, ap);
	len = rc;
    }
    else if (!strcmp(macro, "pid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_PID);
	len = rc;
    }
    else if (!strcmp(macro, "uid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_UID);
	len = rc;
    }
    else if (!strcmp(macro, "euid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_EUID);
	len = rc;
    }
    else if (!strcmp(macro, "suid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_SUID);
	len = rc;
    }
    else if (!strcmp(macro, "gid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_GID);
	len = rc;
    }
    else if (!strcmp(macro, "egid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_EGID);
	len = rc;
    }
    else if (!strcmp(macro, "sgid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_SGID);
	len = rc;
    }
    else if (!strcmp(macro, "comm")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_COMM);
	len = rc;
    }
    else if (!strcmp(macro, "ppid")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_PPID);
	len = rc;
    }
    else if (!strcmp(macro, "pcomm")) {
	rc = tracker_stringify_var(buf, buflen, VT_PROC_PCOMM);
	len = rc;
    }
    else if (!strcmp(macro, "retval")) {
	rc = tracker_stringify_var(buf, buflen, VT_RETVAL);
	len = rc;
    }

    return len;
}

/* log the syscall that matched one of the rules in 'rules' */
static void 
tracker_do_logging(rule_list_t* rules, 
		   int rule_id, 
		   int syscall_id, 
		   const char* syscall_name, 
		   const char* log_format, 
		   const char* param_spec, 
		   va_list ap)
{
    char buf[MAX_LOG_RECORD+1];
    char sbuf[MAX_LOG_RECORD+1];
    const char* pf = log_format;
    char* pb = buf;
    char const * ps;
    int len, mlen;

    while (*pf && pb - buf < MAX_LOG_RECORD) {
	if (*pf == '%') {
	    /* find next semicolon. */
	    if (*++pf) {
		if (*pf == '%') {
		    /* translate '%%' to '%'. */
		    pf++;
		    if (pb - buf < MAX_LOG_RECORD) {
			*pb++ = '%';
			pf++;
			continue;
		    }
		}
		else {
		    /* translate macro name into value. */
		    ps = pf; 
		    /* parse out the macro name - from the current char
		       until the next non alphabetic character */
		    while (isalpha(*ps++))
			;
		    ps--;
		    mlen = sct_min(ps - pf, MAX_LOG_RECORD - (pb - buf));
		    strncpy(sbuf, pf, mlen);
		    sbuf[mlen] = '\0';
		    len = tracker_translate_macro(rules,
					          rule_id,
					          syscall_id,
					          syscall_name,
					          sbuf,
					          pb,
					          MAX_LOG_RECORD - (pb - buf),
					          param_spec,
					          ap);
		    pf = ps;
		    pb += len;
		    continue;
		}
	    }
	}
	else {
	    if (pb - buf < MAX_LOG_RECORD) {
		*pb++ = *pf++;
		continue;
	    }
	}
    }
    *pb = '\0';
    buf[MAX_LOG_RECORD-1] = '\n';
    if (pb > buf && pb-buf < MAX_LOG_RECORD) {
	*(pb-1) = '\n';
	*(pb) = '\0';
    }
    buf[MAX_LOG_RECORD-2] = '.';
    buf[MAX_LOG_RECORD-3] = '.';
    buf[MAX_LOG_RECORD-4] = '.';
    PPRINT(KERN_NOTICE "%s", buf);
}

/* perform an action in the current system call's context. */
/* accepts a system call ID and its (var_args) parameters. */
void
tracker_perform_action(rule_list_t* rules,
		       action_t* action,
		       int rule_id,
		       int syscall_id,
		       const char* syscall_name, 
		       const char* param_spec,
		       ...)
{
    if (action) {
	/* TODO TODO - handle other action types. */
	/* also handle logging with varargs.      */
	switch (action->type) {
	case AT_LOG:
	{
	    const char* log_format = action->u.log.format;
	    va_list ap;
	    va_start (ap, param_spec);

	    if (!log_format || !*log_format)
		log_format =
		    "syscall: %pid[%comm]: %sid_%sname(%params) (rule %ruleid)\n";
	    tracker_do_logging(rules,
			       rule_id,
			       syscall_id,
			       syscall_name,
			       log_format,
			       param_spec,
			       ap);
	    va_end(ap);
	    break;
	}		
	default:
	    PPRINT(KERN_NOTICE "syscall '%s(%d)' invoked - unknown action %d\n",
		   syscall_name, syscall_id, action->type);
	    break;
	}
    }
}

