/*
 *$Id: rule_engine_public.h,v 1.15 2001/12/19 07:57:21 amirs Exp $
 */
#ifndef _RULE_ENGINE_PUBLIC_H
#define _RULE_ENGINE_PUBLIC_H

/*
 * Public interface of the rules matching engine.
 */


/*
 * forward declerations of data types used by the tracking module.
 */

#ifdef __cplusplus 
extern "C" {
#endif

/* values used for booleans. */
#define	TF_TRUE		1
#define	TF_FALSE	0

/* opaque tracker filter structure. */
typedef struct tracker_filter_node filter_node_t;

/* opaque tracker action structure. */
typedef struct tracker_action action_t;

/* opaque tracker rules list structure. */
typedef struct rule_list rule_list_t;

/* enum containing supported variable types. */
typedef enum {
	VT_UNDEF = 0,
	VT_SYSCALL_ID,
	VT_PROC_UID,
	VT_PROC_EUID,
	VT_PROC_SUID,
	VT_PROC_GID,
	VT_PROC_EGID,
	VT_PROC_SGID,
	VT_PROC_PID,
	VT_PROC_COMM,
	VT_RETVAL,
	VT_PROC_PPID,
	VT_PROC_PCOMM,
	VT_PROC_ANCESTORS,
	VT_PARAM,	/* indexed variable type. */
	VT_LAST_TYPE
} var_type_t;

/* types of supported data constants. */
typedef enum {
	DT_BOOL = 0,
	DT_CHAR,
	DT_INT,
	DT_UINT,
	DT_SHORT,
	DT_USHORT,
	DT_LONG,
	DT_ULONG,
	DT_FLOAT,
	DT_STRING,
	DT_MULTI,
	DT_LAST_TYPE
} data_type_t;

/* types of expression syntax/semantic errors. */
typedef enum {
	ERR_NONE = 0,
	ERR_INVALID_DATA_TYPE,
	ERR_INVALID_NODE,
	ERR_INVALID_OPERATOR,
	ERR_TYPE_MISMATCH,
	ERR_BUF_TOO_SMALL,
	ERR_INVALID_VAR
} exp_error_t;

/* types of rules - 'before' or 'after'. */
typedef enum {
	RW_BEFORE = 0,
	RW_AFTER
} rule_when_t;

/*
 * functions that are used to build a filtering expression.
 */

/* create primitive types - int, string, variable, indexed variable */
/* (e.g. parameter).                                                */
filter_node_t* make_bool_val(int num);
filter_node_t* make_int_val(int num);
filter_node_t* make_uint_val(unsigned int num);
filter_node_t* make_char_val(char c);
filter_node_t* make_short_val(short num);
filter_node_t* make_ushort_val(unsigned short num);
filter_node_t* make_long_val(long num);
filter_node_t* make_ulong_val(unsigned long num);
filter_node_t* make_float_val(float num);
filter_node_t* make_string_val(const char* str);
					/* str is copied into the node. */
filter_node_t* make_string_val_without_copy(char* str);
					/* str is used in the node. */
/* make a filter node representing a non-indexed variable reference. */
filter_node_t* make_var_val(var_type_t v, data_type_t var_data_type);
/* make a filter node representing a non-indexed variable's field reference. */
filter_node_t* make_var_field_val(var_type_t v,
				  int field_index,
				  data_type_t var_data_type);
/* make a filter node representing an indexed variable reference. */
filter_node_t* make_var_indexed_val(var_type_t v,
				    int var_index,
				    data_type_t var_data_type);
/* make a filter node representing an indexed variable's field reference. */
filter_node_t* make_var_indexed_field_val(var_type_t,
					  int var_index,
					  int field_index,
					  data_type_t field_data_type);

/* create combined expressions using operators.                 */
/* checks for matching (e.g. can't comprate an int and an str). */
/* NOTE - parameters are "swalloed" by the generated node, and  */
/* will be freed when the generated node is freed.              */
filter_node_t* make_equal(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_less_then(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_less_or_equal(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_greater_then(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_greater_or_equal(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_not_equal(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_logical_and(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_logical_or(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_logical_not(filter_node_t* f);
/* these would work only for numeric types. */
filter_node_t* make_bit_and(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_bit_or(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_bit_xor(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_bit_not(filter_node_t* f);
filter_node_t* make_bit_left_shift(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_bit_right_shift(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_math_plus(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_math_minus(filter_node_t* f1, filter_node_t* f2);
filter_node_t* make_in_match(filter_node_t* f1, filter_node_t* f2);
/* these would work only for string types. */
filter_node_t* make_pattern_match(filter_node_t* f1, filter_node_t* f2);

/* vector handling functions. */
filter_node_t* make_filter_vector(int len);
/* inserts the given filter node into the given vector node. */
/* returns '1' on success, '0' if 'vec' is not a vector.     */
/* '-1' if index is out of range.                            */
int insert_filter_node(filter_node_t* vec, int index, filter_node_t* f);
/* return the length of the given vector node. returns '-1'  */
/* if 'vec' is not a vector.				     */
int get_filter_vector_len(filter_node_t* vec);
/* return the filter node in the given index of the given vector. */
/* returns NULL if index is out of range, or vec is not a vector. */
filter_node_t* get_filter_node_at(filter_node_t* vec, int index);

/* free the given filter node, as well as all sub-nodes and */
/* heap memory it contains, directly and indirectly.        */
void free_filter_node(filter_node_t* f);

/* serialize a filter node (and all its sub-nodes) into a */
/* buffer, for export purposes. returns the size of the   */
/* serialized data, 0 if there was an error, -1 if buffer */
/* is too small.                                          */
int serialize_filter_node(filter_node_t* f, unsigned char* buf, int buflen);

/* de-serialize a filter node (and all its sub-nodes)     */
/* from a buffer, for import purposes. returns the new    */
/* allocated filter node (the top of the tree), or NULL   */
/* if an error occured.                                   */
filter_node_t* deserialize_filter_node(unsigned char* buf, int buflen);

/* print a filter node (recursively) to a "standard" device */
void print_filter_node(filter_node_t* f);

/* evaluate the given filter node. */
/* since this is a generic function, it returns a */
/* filter node, containing a constant and the     */
/* proper result.                                 */
exp_error_t eval_filter_node(filter_node_t* f, filter_node_t* result);

/*
 * Functions for actions manipulation.
 */

/* allocate an action object for logging purposes. */
action_t* make_action_log(int priority, const char* log_format);

/* allocate an action object for rule enabling purposes. */
action_t* make_action_enable_rule(int rule_id);

/* allocate an action object for rule disabling purposes. */
action_t* make_action_disable_rule(int rule_id);

/* allocate an action object for process suspending. */
action_t* make_action_suspend_process(void);

/* allocate an action object for process killing. */
action_t* make_action_kill_process(void);

/* allocate an action object for system call failing. */
action_t* make_action_fail_syscall(int error_code);

/* free the action object */
void free_action(action_t* a);

/* serialize an action node into a buffer, for export     */
/* purposes. returns the size of the serialized data, 0   */
/* if there was an error, -1 if buffer is too small.      */
int serialize_action_node(action_t* a, unsigned char* buf, int buflen);

/* de-serialize an action node from a buffer, for import  */
/* purposes. returns the new allocated action node or     */
/* NULL if an error occuerd.                              */
action_t* deserialize_action_node(unsigned char* buf, int buflen);

/* print an action node to a "standard" device */
void print_action_node(action_t* a);


/*
 * Functions for tracking rules manipulation.
 */

/* serialize a tracking rule into a buffer, for export   */
/* purposes. returns the size of the serialized data, 0  */
/* if there was an error, -1 if the buffer is too small. */
int serialize_tracking_rule(int rule_id, rule_when_t when,
			    int syscall_id,
			    action_t* a, filter_node_t* f,
		            unsigned char* buf, int buflen);

/* de-serialize a tracking rule from a buffer, for import */
/* purposes. returns the 1 for success, -1 for failure.   */
/* the given action rule and filter node are filled with  */
/* elements geenrated from the buffer, and should be      */
/* freed by the caller.                                   */
int deserialize_tracking_rule(unsigned char* buf, int buflen, 
			      int* p_rule_id, rule_when_t* p_when,
			      int* p_syscall_id,
			      action_t** pa, filter_node_t** pf);

/* create an empty tracking rules list. */
rule_list_t* create_tracking_rule_list(void);

/* delete the given tracking rule list, freeing any rule found in it. */
void delete_tracking_rule_list(rule_list_t* rules);

/* add a new tracking rule. returns the new rule's id */
/* returns '0' if an error occured, '-1' if there is  */
/* already a rule with the given rule ID.             */
int add_tracking_rule(rule_list_t* rules,
		      int rule_id,
		      filter_node_t* filter,
		      action_t* action);

/* delete a tracking rule. */
void del_tracking_rule(rule_list_t* rules, int rule_id);

/* deletes all tracking rules in the list. */
void del_all_tracking_rules_in_list(rule_list_t* rules);

/* check if there is a rule with the given ID. */
/* returns '1' if there is, '0' otehrwise.     */
int has_tracking_rule(rule_list_t* rules, int rule_id);

/* returns the number of rules in the given rule list. */
int get_tracking_rule_count(rule_list_t* rules);

/* print out the list of current tracking rules. */
void print_tracking_rules(rule_list_t* rules);

/*
 * functions used by tracking functions to perform filtering.
 */

/* filter an incoming system call.                                          */
/* accepts a system call ID and its (var_args) parameters.                  */
/* returns a tracker_action structure if there was a match, NULL otherwise. */
/* on successfull match, the rule id is set into '*rule_id'.                */
action_t* filter_syscall(rule_list_t* rules, int syscall_id, int* rule_id, ...);

/* perform an action in the current system call's context. */
/* accepts a system call ID and its (var_args) parameters. */
void tracker_perform_action(rule_list_t* rules,
			    action_t* action,
			    int rule_id, 
			    int syscall_id, 
			    const char* syscall_name, 
			    const char* param_spec, 
			    ...);

#ifdef __cplusplus 
}
#endif

#endif /* _RULE_ENGINE_PUBLIC_H */
