/*
 * $Id: device_file.c,v 1.5 2001/12/21 19:31:02 mulix Exp $
 */ 

/* 
 * This code is based upon code from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates. 
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h> /* printk() */
#include <linux/slab.h>   /* kmalloc() */
#include <linux/fs.h>     /* everything... */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/proc_fs.h>
#include <linux/fcntl.h>        /* O_ACCMODE */
#include <linux/spinlock.h>     /* spinlock */
#include <asm/uaccess.h>

#include "device_file.h"

#define SCT_MAJOR 0   /* dynamic major by default */

/* the major number we use (0 for dynamic) */
int sct_major =   SCT_MAJOR;

/* the device struct */
sct_dev_t sct_dev;

/* number of current *writers*. only one allowed! */
unsigned int sct_s_count = 0;
spinlock_t sct_s_lock = SPIN_LOCK_UNLOCKED;

/* the data store we use */
static char buffer[1024];

/*
 * Open and close
 */

int sct_open (struct inode *inode, struct file *filp)
{
    unsigned int open_mode = (filp->f_flags & O_ACCMODE);
    if ((open_mode == O_WRONLY) || (open_mode == O_RDWR))
    {
	spin_lock(&sct_s_lock);
	
	if (sct_s_count) {
	    spin_unlock(&sct_s_lock);
	    return -EBUSY; /* already open */
	}
	sct_s_count++;
	spin_unlock(&sct_s_lock);
    }
    
    /* and use filp->private_data to point to the device data */
    if (!filp->private_data)
	filp->private_data = &sct_dev;
    
    MOD_INC_USE_COUNT;
    return 0;          /* success */
}

int sct_release (struct inode *inode, struct file *filp)
{
    unsigned int open_mode = (filp->f_flags & O_ACCMODE);
    if ((open_mode == O_WRONLY) || (open_mode == O_RDWR))
    {
	sct_s_count--;
    }

    MOD_DEC_USE_COUNT;
    return 0;
}


/*
 * Data management: read and write
 */

ssize_t 
sct_write (struct file *filp, const char *buf, size_t count,
	   loff_t *f_pos)
{
    sct_dev_t *dev = filp->private_data;
    ssize_t retval = -ENOMEM; /* our most likely error */

    if (down_interruptible (&dev->sem))
	return -ERESTARTSYS;
    
    if (*f_pos + count > sizeof(buffer))
        count = sizeof(buffer) - *f_pos;

    if (count == 0) {
	retval = -ENOSPC;
	goto nomem;
    }

    if (copy_from_user (buffer, buf, count)) {
        retval = -EFAULT;
        goto nomem;
    }

    *f_pos += count;

/*  we now have some of the user supplied bytestream  */
/*  in the buffer. we need to copy it to a real buffer,  */
/*  and if we got the complete packet, process it.  */

    up (&dev->sem);
    return count;

  nomem:
    up (&dev->sem);
    return retval;
}

ssize_t 
sct_read (struct file *filp, char *buf, size_t count,
	  loff_t *f_pos)
{
    sct_dev_t *dev = filp->private_data; /* the first listitem */
    ssize_t retval = 0;

    if (down_interruptible (&dev->sem))
	return -ERESTARTSYS;

    if (count + *f_pos > sizeof(buffer))
        count = sizeof(buffer) - *f_pos;

    if (copy_to_user (buf, buffer, count)) {
        retval = -EFAULT;
        goto nothing;
    }
    *f_pos += count;

    up (&dev->sem);
    return count;

  nothing:
    up (&dev->sem);
    return retval;
}

/*
 * The fops
 */

struct file_operations sct_fops = {
    read: sct_read,
    write: sct_write, 
    open: sct_open,
    release: sct_release
};
