/*
 *$Id: module_main.c,v 1.24 2001/12/21 19:31:02 mulix Exp $
 */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/modversions.h>
#include <linux/config.h>
#include <linux/unistd.h>
#include <linux/sys.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/sysctl.h>
#include <linux/limits.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/semaphore.h>

#include "sct_rules_private.h"
#include "tracker_sysctl.h"
#include "filter_vars_public.h"
#include "rule_engine_public.h"
#include "syscall_hijack_public.h"
#include "module_interface.h"
#include "device_file.h"
#include "sct_compat.h"


MODULE_AUTHOR("guy keren <choo@actcom.co.il>, mulix <mulix@actcom.co.il>");

MODULE_DESCRIPTION("System Call Filter Evaluation and Userspace Interface, "
		   "for 'syscalltrack'");

MODULE_LICENSE("GPL");


/*
 * Forward declarations.
 */

/* the dispatch function for all sysctl handling code */
int tracker_sysctl(ctl_table *table, int *name, int namelen,
                   void *oldval, size_t *oldlenp,
                   void *newval, size_t newlen, void **context);

/* initialize this mux function for this syscall_id */
void init_mux_function(int syscall_id,
		       int mux_func_id,
		       func_ptr track_before_mux_func,
		       func_ptr track_after_mux_func);

/*
 * Static data structures.
 */

/* structure used to store info about type data for a struct type. */
struct struct_data {
    long* offsets_vec;		/* offsets of struct's fields.		*/
    long num_fields;		/* number of fields in struct.		*/
};

/* defined in sct_rules_private.h */
struct syscall_track_data track_data[NR_syscalls];

/* defined in sct_rules_private.h */
struct mux_data_t mux_data[NR_MUX_SYSCALLS];

/* data used to register sysctl functions for user-mode access to the module. */
static ctl_table tracker_table[] = {
	{CTL_TRACKER_CMD_ADD_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_DEL_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_ENABLE_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_DISABLE_RULE, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES, NULL, NULL, 1024,
	 0600, NULL, NULL, &tracker_sysctl},
	{CTL_TRACKER_CMD_PRINT_ALL_RULES, NULL, NULL, 1024, 
	 0600, NULL, NULL, &tracker_sysctl}, 
	{0}
};

static ctl_table tracker_top_table[] = {
	{CTL_TRACKER, NULL, NULL, 0, 0500, tracker_table},
	{0}
};


static struct ctl_table_header* tracker_sysctl_header = NULL;

#include "syscall_tracker_autogen.c"

/* build a rule from this buffer. */
/* return negative error on error, 0 for success */
int tracker_build_rule(unsigned char* buf, int len, int* rule_id, 
		       rule_when_t* when, int* syscall_id, 
		       int* mux_func_id, action_t** act, 
		       filter_node_t** f)
{
    int rc = deserialize_tracking_rule(buf, len, rule_id, when,
				       syscall_id, act, f);
    if (rc <= 0) {
	printk(KERN_NOTICE "tracker_build_rule: failed to deserialize tracking rule, "
	       "err '%d'\n", rc);
	return -EINVAL; /* todo: anything better? */
    }
    
    *mux_func_id = FUNC_PART_OF_SYSCALL_ID(*syscall_id);
    *syscall_id = SYSCALL_PART_OF_SYSCALL_ID(*syscall_id);
    
    return rc;
}

/* sanity checks */
int sanity_check_user_input(int syscall_id, int mux_func_id, 
			    rule_when_t when)
{
    int mux_id = mux_data_index(syscall_id);
    
    if (syscall_id < 0 || syscall_id > NR_syscalls) {
        printk(KERN_NOTICE "sanity_check_user_input: syscall id "
	       "'%d; out of range\n", syscall_id);
	return -EINVAL;
    }
    if (mux_func_id != 0) { /* make these checks only for mux functions */

	if ((mux_func_id < 0) || /* there cannot be such a mux function */
	    (mux_id < 0) || /* or this syscall is not a mux syscall */
	    /* or it is a mux syscall, but without this mux function */
	    (mux_func_id >= mux_data[mux_id].num_muxed)) { 
	    printk(KERN_NOTICE "syscall id %d , mux function id %d out of range\n",
		   syscall_id, mux_func_id);
	    return -EINVAL;
	}
	if (mux_func_id > 0 &&
	    ((when == RW_BEFORE && !mux_data[mux_id].before_rules[mux_func_id])
	     ||
	     (when == RW_AFTER && !mux_data[mux_id].after_rules[mux_func_id]))) {
	    printk(KERN_NOTICE "syscall id %d , mux function id %d not supported\n",
		   syscall_id, mux_func_id);
	    return -EINVAL;
	}
    }
    return 0; /* success */
}

/* sysctl handling routine. allows adding/deleting/enabling/disabling rules. */
int tracker_sysctl(ctl_table *table,
		   int *name, int namelen,
                   void *oldval, size_t *oldlenp,
                   void *newval, size_t newlen, void **context)
{
    int i;
    int rule_id;
    rule_when_t when;
    int syscall_id;
    action_t* a;
    filter_node_t* f;
    int retval = 0;
    unsigned char* buf = NULL;
    int mux_func_id = 0;

    printk(KERN_DEBUG "tracker_sysctl: namelen = '%d', newlen = '%d'\n", namelen, newlen);
    for (i=0; i<namelen; i++) {
	printk(KERN_DEBUG "tracker_sysctl: name[%d] = %d\n", i, name[i]);
    }
    if (newlen <= 0 || namelen != 1) {
	retval = -EINVAL;
	goto err_out;
    }

    buf = kmalloc(newlen, GFP_KERNEL);
    if (!buf){
	retval = -ENOMEM;
	printk(KERN_NOTICE "tracker_sysctl: buffer allocation failed\n");
	goto err_oom;
    }
    if (copy_from_user(buf, newval, newlen)) {
	retval = -EFAULT; /* TODO: enomem? eaccess? */
	printk(KERN_NOTICE "tracker_sysctl: failed to copy from user\n");
	goto err_copy_from_user;
    }
    printk(KERN_NOTICE "tracker_sysctl: copied data from user\n");

    /* build a rule from the user supplied buffer */
    if ((retval = tracker_build_rule(buf, newlen, &rule_id, &when, 
				     &syscall_id, &mux_func_id, 
				     &a, &f)) < 0)
	goto err_build_rule; 
    
    if ((retval = sanity_check_user_input(syscall_id, mux_func_id, 
					  when)) < 0)
	goto err_failed_sanity_check;

    /* which operation did the user request? */
    switch (name[0]) {
        case(CTL_TRACKER_CMD_ADD_RULE):
	    retval = tracker_sysctl_add_rule(rule_id, when,
					     syscall_id, mux_func_id,
					     f, a);
	    break;
	
        case(CTL_TRACKER_CMD_DEL_RULE):
	    retval = tracker_sysctl_del_rule(rule_id, when,
					     syscall_id, mux_func_id);
	    break;
	
	case(CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL):
	    retval = tracker_sysctl_del_all_rules_for_syscall(syscall_id,
							      mux_func_id);
	    break;

        case(CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES):
	    retval = tracker_sysctl_del_all_module_rules();
 	    break;

        case(CTL_TRACKER_CMD_PRINT_ALL_RULES):
	    retval = tracker_sysctl_print_all_rules();
	    break;

        default:
	    printk(KERN_NOTICE "tracker_sysctl: invalid or unsupported sysctl "
		   "operation %d\n", name[0]);
	    retval = -EINVAL;
    }
    /* all is well - we can return peacefully. */

 err_failed_sanity_check:
 err_build_rule:
 err_copy_from_user:
    kfree(buf);
 err_oom:
 err_out:

    return retval;
}

/* init muxed syscalls data, for the given syscall id. */
void
init_mux_syscall(int syscall_id,
		 func_ptr syscall_track_before_func,
		 func_ptr syscall_track_after_func,
		 int num_muxed_functions)
{
    int j, id;

    if ((id = mux_data_index(syscall_id)) < 0) {
	printk(KERN_WARNING "init_mux_syscall(): syscall '%d' is "
	       "NOT s mux syscall!\n", syscall_id);
	return;
    }

    before_track_functions[syscall_id] = syscall_track_before_func;
    after_track_functions[syscall_id] = syscall_track_after_func;
    mux_data[id].mux_tracker = syscall_track_before_func;
    mux_data[id].num_muxed = num_muxed_functions;

    mux_data[id].muxed_functions =
	(func_ptr*)kmalloc(num_muxed_functions*sizeof(func_ptr), GFP_KERNEL);
    if (mux_data[id].muxed_functions == NULL)
	goto init_mux_syscall_cleanup_1;
    
    mux_data[id].before_rules =
    	(rule_list_t**)kmalloc(num_muxed_functions*sizeof(rule_list_t*),
			       GFP_KERNEL);
    if (mux_data[id].before_rules == NULL)
	goto init_mux_syscall_cleanup_2;

    mux_data[id].after_rules =
    	(rule_list_t**)kmalloc(num_muxed_functions*sizeof(rule_list_t*),
			       GFP_KERNEL);
    if (mux_data[id].before_rules == NULL)
	goto init_mux_syscall_cleanup_3;

    for (j=0; j < num_muxed_functions; j++) {
        mux_data[id].muxed_functions[j] = (func_ptr)NULL;
        mux_data[id].before_rules[j] = (rule_list_t*)NULL;
        mux_data[id].after_rules[j] = (rule_list_t*)NULL;
    }

    /* success, we can return now */
    return;

 init_mux_syscall_cleanup_3:
    kfree(mux_data[id].before_rules);

 init_mux_syscall_cleanup_2:
    kfree(mux_data[id].muxed_functions);

 init_mux_syscall_cleanup_1:
    before_track_functions[syscall_id] = (func_ptr)NULL;
    after_track_functions[syscall_id] = (func_ptr)NULL;
    mux_data[id].mux_tracker = (func_ptr)NULL;
    mux_data[id].num_muxed = 0;
}

void
init_mux_function(int syscall_id,
		  int mux_func_id,
		  func_ptr track_before_mux_func,
		  func_ptr track_after_mux_func)
{
    int id;

    if ((id = mux_data_index(syscall_id)) < 0) {
	printk(KERN_WARNING "init_mux_function(): syscall '%d' is "
	       "NOT s mux syscall!\n", syscall_id);
	return;
    }

    mux_data[id].muxed_functions[mux_func_id] = track_before_mux_func;
    mux_data[id].before_rules[mux_func_id] = create_tracking_rule_list();
    mux_data[id].after_rules[mux_func_id] = create_tracking_rule_list();
}

void init_callback_functions(void) {
	reset_all_vars();
	set_persistent_callback(VT_PROC_UID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_EUID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_SUID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_GID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_EGID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_SGID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_PID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_COMM, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_PPID, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_PCOMM, proc_current_callback, NULL);
	set_persistent_callback(VT_PROC_ANCESTORS, proc_current_callback, NULL);
}

/* forward referenec for a special function to handle 'socketcall' syscall, 'before' rules. */
extern asmlinkage long sct_sys_socketcall(int call, unsigned long *args);

/* forward referenec for a special function to handle 'socketcall' syscall, 'after' rules. */
extern asmlinkage long sct_sys_socketcall_after(int syscall_rc, int call, unsigned long *args);

int
init_module (void)
{
	int i, result;

	for (i=0; i<NR_syscalls; ++i) {
		track_data[i].before_rules = create_tracking_rule_list();
		track_data[i].after_rules = create_tracking_rule_list();
	}

	for (i=0; i<NR_MUX_SYSCALLS; ++i) {
	    mux_data[i].mux_tracker = (func_ptr)NULL;
	    mux_data[i].num_muxed = 0;
	    mux_data[i].muxed_functions = (func_ptr*)NULL;
	    mux_data[i].demux_functions = (func_ptr*)NULL;
	    mux_data[i].before_rules = (rule_list_t**)NULL;
	    mux_data[i].after_rules = (rule_list_t**)NULL;
	}

	/* special handling for 'socketcall' syscall - */
	/* register a special tracker function.        */
	init_mux_syscall(102, 
			 (func_ptr)sct_sys_socketcall, 
			 (func_ptr)sct_sys_socketcall_after, 
			 18);

	init_mux_functions();
		
	/* init calculation of data used to handle struct syscall parameters. */
	init_struct_offsets();

	init_callback_functions();

	/* register sysctl handling function. */
 	tracker_sysctl_header = register_sysctl_table(tracker_top_table, 0);
	printk(KERN_DEBUG "tracker_sysctl_header - %p\n", tracker_sysctl_header); 

	/* register device file handling functions */
	SET_MODULE_OWNER(&sct_fops);

	/*
	 * Register your major, and accept a dynamic number
	 */
	result = register_chrdev(sct_major, "sct_rules", &sct_fops);
	if (result < 0) {
	    if (tracker_sysctl_header)
		unregister_sysctl_table(tracker_sysctl_header);
	    return result;
	}
	
	if (sct_major == 0) 
	    sct_major = result; /* dynamic */
	
	sema_init(&sct_dev.sem, 1);

	return 0;
}

void
cleanup_module (void)
{
	int i;

        printk(KERN_NOTICE "sct_rules.o: cleaning up syscall filter module\n");

	/* register sysctl handling function. */
	if (tracker_sysctl_header)
		unregister_sysctl_table(tracker_sysctl_header);

	for (i=0; i<NR_syscalls; ++i) {
		delete_tracking_rule_list(track_data[i].before_rules);
		delete_tracking_rule_list(track_data[i].after_rules);
	}

	for (i=0; i<NR_MUX_SYSCALLS; ++i) {
	    mux_data[i].mux_tracker = (func_ptr)NULL;
	    mux_data[i].num_muxed = 0;
	    if (mux_data[i].muxed_functions)
		kfree(mux_data[i].muxed_functions);
	    if (mux_data[i].demux_functions)
		kfree(mux_data[i].demux_functions);
	    if (mux_data[i].before_rules) {
		int j;
		for (j=0; j<mux_data[i].num_muxed; j++)
		    delete_tracking_rule_list(mux_data[i].before_rules[j]);
		kfree(mux_data[i].before_rules);
	    }
	    if (mux_data[i].after_rules) {
		int j;
		for (j=0; j<mux_data[i].num_muxed; j++)
		    delete_tracking_rule_list(mux_data[i].after_rules[j]);
		kfree(mux_data[i].after_rules);
	    }
	}

	unregister_chrdev(sct_major, "sct_rules");
}



/***********************************************************************/    
/***********************************************************************/    

/* special handlign for the 'socketcall' syscall, and its sub-elements. */

/* special function, used to track 'socketcall'. */
/* Argument list sizes for sys_socketcall */
#define AL(x) ((x) * sizeof(unsigned long))
static unsigned char nargs[18]={AL(0),AL(3),AL(3),AL(3),AL(2),AL(3),
				AL(3),AL(3),AL(4),AL(4),AL(4),AL(6),
				AL(6),AL(2),AL(5),AL(5),AL(3),AL(3)};
#undef AL


/*
 *	System call vectors. 
 *
 *	Argument checking cleaned up. Saved 20% in size.
 *  This function doesn't need to set the kernel lock because
 *  it is set by the callees. 
 */

#if 0
asmlinkage long sct_sys_socketpair(int family, int type, int protocol, int usockvec[2])
asmlinkage long sct_sys_bind(int fd, struct sockaddr *umyaddr, int addrlen)
asmlinkage long sct_sys_listen(int fd, int backlog)
asmlinkage long sct_sys_accept(int fd, struct sockaddr *upeer_sockaddr, int *upeer_addrlen)
asmlinkage long sct_sys_connect(int fd, struct sockaddr *uservaddr, int addrlen)
asmlinkage long sct_sys_getsockname(int fd, struct sockaddr *usockaddr, int *usockaddr_len)
asmlinkage long sct_sys_getpeername(int fd, struct sockaddr *usockaddr, int *usockaddr_len)
asmlinkage long sct_sys_sendto(int fd, void * buff, size_t len, unsigned flags,
			   struct sockaddr *addr, int addr_len)
asmlinkage long sct_sys_send(int fd, void * buff, size_t len, unsigned flags)
asmlinkage long sct_sys_recvfrom(int fd, void * ubuf, size_t size, unsigned flags,
			     struct sockaddr *addr, int *addr_len)
asmlinkage long sct_sys_recv(int fd, void * ubuf, size_t size, unsigned flags)
asmlinkage long sct_sys_setsockopt(int fd, int level, int optname, char *optval, int optlen)
asmlinkage long sct_sys_getsockopt(int fd, int level, int optname, char *optval, int *optlen)
asmlinkage long sct_sys_shutdown(int fd, int how)
asmlinkage long sct_sys_sendmsg(int fd, struct msghdr *msg, unsigned flags)
asmlinkage long sct_sys_recvmsg(int fd, struct msghdr *msg, unsigned int flags)
#endif /* 0 */


asmlinkage long sct_sys_socketcall(int call, unsigned long *args)
{
        /*printk(KERN_INFO "sct_sys_socketcall - call '%d'\n", call);*/

	unsigned long a[6];
	unsigned long a0,a1;
	int err;

	if(call<1||call>SYS_RECVMSG)
		return -EINVAL;

	/* copy_from_user should be SMP safe. */
	if (copy_from_user(a, args, nargs[call]))
		return -EFAULT;
		
	a0=a[0];
	a1=a[1];
	
	switch(call) 
	{
		case SYS_SOCKET:
			err = track_before_syscall_102_1(a0,a1,a[2]);
			break;
		case SYS_BIND:
			err = track_before_syscall_102_2(a0,(struct sockaddr *)a1, a[2]);
			break;
		case SYS_CONNECT:
			err = track_before_syscall_102_3(a0, (struct sockaddr *)a1, a[2]);
			break;
		case SYS_LISTEN:
			err = track_before_syscall_102_4(a0,a1);
			break;
		case SYS_ACCEPT:
			err = track_before_syscall_102_5(a0,(struct sockaddr *)a1, (int *)a[2]);
			break;
		case SYS_SHUTDOWN:
			err = track_before_syscall_102_13(a0,a1);
			break;
		default:
			err = 0;
			break;
#if 0
		case SYS_GETSOCKNAME:
			err = demux_sys_getsockname(a0,(struct sockaddr *)a1, (int *)a[2]);
			break;
		case SYS_GETPEERNAME:
			err = demux_sys_getpeername(a0, (struct sockaddr *)a1, (int *)a[2]);
			break;
		case SYS_SOCKETPAIR:
			err = demux_sys_socketpair(a0,a1, a[2], (int *)a[3]);
			break;
		case SYS_SEND:
			err = demux_sys_send(a0, (void *)a1, a[2], a[3]);
			break;
		case SYS_SENDTO:
			err = demux_sys_sendto(a0,(void *)a1, a[2], a[3],
					 (struct sockaddr *)a[4], a[5]);
			break;
		case SYS_RECV:
			err = demux_sys_recv(a0, (void *)a1, a[2], a[3]);
			break;
		case SYS_RECVFROM:
			err = demux_sys_recvfrom(a0, (void *)a1, a[2], a[3],
					   (struct sockaddr *)a[4], (int *)a[5]);
			break;
		case SYS_SETSOCKOPT:
			err = demux_sys_setsockopt(a0, a1, a[2], (char *)a[3], a[4]);
			break;
		case SYS_GETSOCKOPT:
			err = demux_sys_getsockopt(a0, a1, a[2], (char *)a[3], (int *)a[4]);
			break;
		case SYS_SENDMSG:
			err = demux_sys_sendmsg(a0, (struct msghdr *) a1, a[2]);
			break;
		case SYS_RECVMSG:
			err = demux_sys_recvmsg(a0, (struct msghdr *) a1, a[2]);
			break;
		default:
			err = -EINVAL;
			break;
#endif /* 0 */
	}
	return err;
}

asmlinkage long sct_sys_socketcall_after(int syscall_rc, int call, unsigned long *args)
{
	unsigned long a[6];
	unsigned long a0,a1;
	int err;

	/*printk(KERN_INFO "sct_sys_socketcall_after - call '%d'\n", call);*/
	
	if(call<1||call>SYS_RECVMSG)
		return -EINVAL;

	/* copy_from_user should be SMP safe. */
	if (copy_from_user(a, args, nargs[call]))
		return -EFAULT;
		
	a0=a[0];
	a1=a[1];
	
	switch(call) 
	{
		case SYS_SOCKET:
			err = track_after_syscall_102_1(syscall_rc, a0,a1,a[2]);
			break;
		case SYS_BIND:
			err = track_after_syscall_102_2(syscall_rc, a0,(struct sockaddr *)a1, a[2]);
			break;
		case SYS_CONNECT:
			err = track_after_syscall_102_3(syscall_rc, a0, (struct sockaddr *)a1, a[2]);
			break;
		case SYS_LISTEN:
			err = track_after_syscall_102_4(syscall_rc, a0,a1);
			break;
		case SYS_ACCEPT:
			err = track_after_syscall_102_5(syscall_rc, a0,(struct sockaddr *)a1, (int *)a[2]);
			break;
		case SYS_SHUTDOWN:
			err = track_after_syscall_102_13(syscall_rc, a0,a1);
			break;
		default:
			err = 0;
			break;
#if 0
		case SYS_GETSOCKNAME:
			err = demux_sys_getsockname(a0,(struct sockaddr *)a1, (int *)a[2]);
			break;
		case SYS_GETPEERNAME:
			err = demux_sys_getpeername(a0, (struct sockaddr *)a1, (int *)a[2]);
			break;
		case SYS_SOCKETPAIR:
			err = demux_sys_socketpair(a0,a1, a[2], (int *)a[3]);
			break;
		case SYS_SEND:
			err = demux_sys_send(a0, (void *)a1, a[2], a[3]);
			break;
		case SYS_SENDTO:
			err = demux_sys_sendto(a0,(void *)a1, a[2], a[3],
					 (struct sockaddr *)a[4], a[5]);
			break;
		case SYS_RECV:
			err = demux_sys_recv(a0, (void *)a1, a[2], a[3]);
			break;
		case SYS_RECVFROM:
			err = demux_sys_recvfrom(a0, (void *)a1, a[2], a[3],
					   (struct sockaddr *)a[4], (int *)a[5]);
			break;
		case SYS_SETSOCKOPT:
			err = demux_sys_setsockopt(a0, a1, a[2], (char *)a[3], a[4]);
			break;
		case SYS_GETSOCKOPT:
			err = demux_sys_getsockopt(a0, a1, a[2], (char *)a[3], (int *)a[4]);
			break;
		case SYS_SENDMSG:
			err = demux_sys_sendmsg(a0, (struct msghdr *) a1, a[2]);
			break;
		case SYS_RECVMSG:
			err = demux_sys_recvmsg(a0, (struct msghdr *) a1, a[2]);
			break;
		default:
			err = -EINVAL;
			break;
#endif /* 0 */
	}
	return err;
}


/* end of special section for dealing with socketcall. */
/************************************************************************/
/************************************************************************/
