/*
 * $Id: update_rules.c,v 1.1 2001/12/21 19:31:02 mulix Exp $
 */ 

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/errno.h>
#include <asm/semaphore.h>

#include "rule_engine_public.h"
#include "sct_rules_private.h"

#include "update_rules.h"

/* utility functions */

static inline int
syscall_get_rule_count_when(int syscall_id, rule_when_t when)
{
    int count = 0;
    int id, i; 

    /* special handling for mux syscalls. */
    if ((id = mux_data_index(syscall_id)) >= 0) {
	struct mux_data_t* data = &mux_data[id];
	for (i = 0; i < data->num_muxed; ++i) {
	    count += ((when == RW_BEFORE) ?
		get_tracking_rule_count(data->before_rules[i]) :
		get_tracking_rule_count(data->after_rules[i]));
	}
    } else { /* regular syscalls */
	count = ((when == RW_BEFORE) ?
		 get_tracking_rule_count(track_data[syscall_id].before_rules) :
		 get_tracking_rule_count(track_data[syscall_id].after_rules));
    }
    return count;
}

static inline int 
syscall_is_hijacked_when(int syscall_id, rule_when_t when)
{
    return (syscall_get_rule_count_when(syscall_id, when) > 0);
}


static inline int 
syscall_is_hijacked(int syscall_id)
{
    return (syscall_is_hijacked_when(syscall_id, RW_BEFORE) ||
	    syscall_is_hijacked_when(syscall_id, RW_AFTER));
}

/* which rule list will we be working on? this depends on whether this   */
/* is a mux function or a regular function, and whether this is a before */
/* or after rule                                                         */
static inline rule_list_t*
syscall_get_rule_list(int syscall_id, int mux_func_id, rule_when_t when)
{
    rule_list_t* p_rules = NULL;   /* assume failure. */
    int mux_idx;

    if (mux_func_id == 0) { /* regular function, not muxed */
	p_rules = ((when == RW_BEFORE) ? 
		   track_data[syscall_id].before_rules :
		   track_data[syscall_id].after_rules);
    } else {
	if ((mux_idx = mux_data_index(syscall_id)) >= 0) {
	    p_rules = ((when == RW_BEFORE) ? 
		       mux_data[mux_idx].before_rules[mux_func_id] :
		       mux_data[mux_idx].after_rules[mux_func_id]);
	}
    }

    return p_rules;
}



/* adds a new rule for a given system call. */
int tracker_add_rule(int rule_id,
		     rule_when_t when,
		     int syscall_id,
		     int mux_func_id,
		     filter_node_t* f,
		     action_t* a)
{
    int rc = 0;
    int retval = 0;
    rule_list_t* p_rules;
    const char* when_str = (when == RW_BEFORE ? "before" : "after");
    int when_not = (when == RW_BEFORE ? RW_AFTER : RW_BEFORE);
    
    printk(KERN_NOTICE "adding tracking rule %d (%s) for syscall %d, mux_func_id %d\n",
	   rule_id, when_str, syscall_id, mux_func_id);
    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -EINTR;
        goto err_add_rule_lock;
    }
    /* which rule list will we be working on? this depends on whether this is 
       a mux function or a regular function, and whether this is a before
       or after rule */
    p_rules = syscall_get_rule_list(syscall_id, mux_func_id, when);
    if (!p_rules)
	goto err_add_rule_no_rule_list;

    rc = add_tracking_rule(p_rules, rule_id, f, a);
    if (rc < 0) {
	printk (KERN_NOTICE "add tracking_rule returned '%d'\n", rc);
        retval = -ENOMEM; /* todo: something better? */
        goto err_add_rule_add_rule;
    }

    /* make sure we don't release 'f', since it was consumed */
    /* by the 'add_tracking_rule' function.                  */
    f = NULL;
    free_action(a);
    a = NULL;
    print_tracking_rules(p_rules);
    
    /* if it's a first rule, need to hijack */
    if (syscall_get_rule_count_when(syscall_id, when) == 1) {
	printk(KERN_NOTICE "first '%s' rule for syscall - hijacking syscall\n",
	       when_str);
	if (when == RW_BEFORE)
	    rc = hijack_syscall_before(
				syscall_id,
				(func_ptr)before_track_functions[syscall_id]);
	else
	    rc = hijack_syscall_after(
			        syscall_id,
				(func_ptr)after_track_functions[syscall_id]);
	printk(KERN_NOTICE "hijack_syscall_%s returned '%d'\n", when_str, rc);
	if (rc < 0) {
	    printk(KERN_NOTICE "hijack_syscall_%s failed, error '%d'\n",
		   when_str, rc);
	    retval = (rc != -ERESTARTSYS ? rc : -EINTR);
	    /* remove rule */
	    del_tracking_rule(p_rules, rule_id);
	    goto err_add_rule_hijack;
	}
	/* if it's the first hijack */
	if (!syscall_is_hijacked_when(syscall_id, when_not)) {
	    MOD_INC_USE_COUNT;
	}
    }

 err_add_rule_hijack:
 err_add_rule_add_rule:
 err_add_rule_no_rule_list:
    up(&tracker_sem);
 err_add_rule_lock:
    if (f)
	free_filter_node(f);
    if (a)
	free_action(a);
    
    return retval;
}

/* deletes an existing rule for a given system call. */
int
tracker_del_rule(int rule_id,
		 rule_when_t when,
		 int syscall_id,
		 int mux_func_id)
{
    int rc;
    int retval = 0;
    rule_list_t* p_rules;
    const char* when_str = (when == RW_BEFORE ? "before" : "after");
    int when_not = (when == RW_BEFORE ? RW_AFTER : RW_BEFORE);

    printk(KERN_NOTICE "tracker_sysctl: deleting tracking rule %d (%s) for syscall %d\n",
           rule_id, when_str, syscall_id);

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_rule_lock;
    }

    p_rules = syscall_get_rule_list(syscall_id, mux_func_id, when);
    if (!p_rules)
	goto err_del_rule_no_rule_list;

    rc = has_tracking_rule(p_rules, rule_id);
    if (!rc) {
	printk(KERN_NOTICE "tracker_sysctl: rule '%d' not found.\n", rule_id);
	goto err_del_rule_no_such_rule;
    }
    
    del_tracking_rule(p_rules, rule_id);
    print_tracking_rules(p_rules);

    /* if it's a the last rule, need to release */
    if (!syscall_is_hijacked_when(syscall_id, when)) {
	printk(KERN_NOTICE "last '%s' rule for syscall - releasing syscall\n",
	       when_str);
	if (when == RW_BEFORE)
	    rc = release_syscall_before(syscall_id);
	else
	    rc = release_syscall_after(syscall_id);
	if (rc < 0) {
	    printk(KERN_NOTICE "release_syscall_%s failed, error '%d'\n",
		   when_str, rc);
	    retval = (rc != -ERESTARTSYS ? rc : -EINTR);
	    goto err_del_rule_release;
	}
	if (!syscall_is_hijacked_when(syscall_id, when_not)) {
	    MOD_DEC_USE_COUNT;
	}
    }

 err_del_rule_release:
 err_del_rule_no_such_rule:
 err_del_rule_no_rule_list:
    up(&tracker_sem);
 err_del_rule_lock:
    
    return retval;
}

/* deletes all existing rules for a given system call. */
/* TODO - handle muxed functions as if they were real syscalls. */
int
tracker_del_all_rules_for_syscall(int syscall_id, int mux_func_id)
{
    int rc1 = 0;
    int rc2 = 0;
    int retval = 0;
    int had_before_rules = 0;
    int had_after_rules = 0;

    printk(KERN_NOTICE "tracker_sysctl: deleting all tracking rules for syscall %d\n",
           syscall_id);

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_all_rules_for_syscall_lock;
    }

    if (!syscall_is_hijacked(syscall_id))
	goto ret_ok_del_all_rules_for_syscall_lock;

    if (track_data[syscall_id].before_rules){
	del_all_tracking_rules_in_list(track_data[syscall_id].before_rules);
	print_tracking_rules(track_data[syscall_id].before_rules);
        had_before_rules = 1;
    }
    if (track_data[syscall_id].after_rules){
	del_all_tracking_rules_in_list(track_data[syscall_id].after_rules);
	print_tracking_rules(track_data[syscall_id].after_rules);
        had_after_rules = 1;
    }
    printk(KERN_NOTICE "tracker_sysctl: del all rules for syscall - releasing syscall\n");
    if (had_before_rules)
        rc1 = release_syscall_before(syscall_id);
    if (had_after_rules)
        rc2 = release_syscall_after(syscall_id);
    if (rc1 < 0) {
        printk(KERN_NOTICE "tracker_sysctl: release_syscall_before failed, error '%d'\n", rc1); 
        retval = (rc1 != -ERESTARTSYS ? rc1 : -EINTR);
        goto err_del_all_rules_for_syscall_release;
    }
    if (rc2 < 0) {
        printk(KERN_NOTICE "tracker_sysctl: release_syscall_after failed, error '%d'\n", rc2); 
        retval = (rc2 != -ERESTARTSYS ? rc2 : -EINTR);
        goto err_del_all_rules_for_syscall_release;
    }

    MOD_DEC_USE_COUNT;
    
 err_del_all_rules_for_syscall_release:
 ret_ok_del_all_rules_for_syscall_lock:
    up(&tracker_sem);
 err_del_all_rules_for_syscall_lock:

    return retval;
}

/* deletes all existing rules in the module */
int
tracker_del_all_module_rules(void)
{
    int rc1 = 0;
    int rc2 = 0;
    unsigned int i;
    int retval = 0;

    printk(KERN_NOTICE "tracker_sysctl: deleting all module tracking rules\n");

    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_del_all_module_rules_lock;
    }

    /* iterate over all syscalls and release the hijacked ones */
    for (i = 0; i < NR_syscalls; ++i){
	int mux_idx;
	/* not hijacked? go on */
	if (!syscall_is_hijacked(i))
	    continue;

	if (track_data[i].before_rules)
	    del_all_tracking_rules_in_list(track_data[i].before_rules);
	if (track_data[i].after_rules)
	    del_all_tracking_rules_in_list(track_data[i].after_rules);
	/* handle muxed syscalls. */
	if ((mux_idx = mux_data_index(i)) >= 0) {
	    struct mux_data_t* data = &mux_data[mux_idx];
	    if (data->num_muxed > 0 && data->before_rules) {
		int j;
		for (j=0; j < data->num_muxed; j++) {
		    if (data->before_rules[j])
			del_all_tracking_rules_in_list(data->before_rules[j]);
		}
	    }
	    if (data->num_muxed > 0 && data->after_rules) {
		int j;
		for (j=0; j < data->num_muxed; j++) {
		    if (data->after_rules[j])
			del_all_tracking_rules_in_list(data->after_rules[j]);
		}
	    }
	}
	
	rc1 = release_syscall_before(i);
	rc2 = release_syscall_after(i);
	if (rc1 < 0) {
	    printk(KERN_NOTICE "tracker_sysctl: release_syscall for syscall '%d' failed, "
		   "error '%d'\n", i, rc1); 
	    retval = (rc1 != -ERESTARTSYS ? rc1 : -EINTR);
	    goto err_del_all_module_rules_release;
	}
	if (rc2 < 0) {
	    printk(KERN_NOTICE "tracker_sysctl: release_syscall for syscall '%d' failed, "
		   "error '%d'\n", i, rc2); 
	    retval = (rc2 != -ERESTARTSYS ? rc2 : -EINTR);
	    goto err_del_all_module_rules_release;
	}
	MOD_DEC_USE_COUNT;
    }
    
 err_del_all_module_rules_release:
    up(&tracker_sem);
 err_del_all_module_rules_lock:
    
    return retval;
}

/* prints all existing rules in the module */
int tracker_print_all_rules(void)
{
    unsigned int i;
    int retval = 0;
    int printed = 0;
    
    /* lock to avoid races with other invocations. */
    if (down_interruptible(&tracker_sem)) {
        retval = -ERESTARTSYS;
        goto err_print_module_rules_lock;
    }
    
    printk (KERN_NOTICE "**************************************\n");
    printk (KERN_NOTICE "syscalltrack - printing defined rules\n");
    for (i = 0; i < NR_syscalls; ++i)
    {
	if (syscall_is_hijacked(i)){
	    int got_any_before = 0;
	    int mux_idx;

	    if (track_data[i].before_rules) {
	        printk(KERN_NOTICE "rules for syscall '%d':\n", i);
	        printk(KERN_NOTICE " 'before' rules:\n");
	        print_tracking_rules(track_data[i].before_rules);
	        printed = 1;
	        got_any_before = 1;
	    }
	    if (track_data[i].after_rules) {
		if (!got_any_before)
	            printk(KERN_NOTICE "rules for syscall '%d':\n", i);
	        printk(KERN_NOTICE " 'after' rules:\n");
	        print_tracking_rules(track_data[i].after_rules);
	        printed = 1;
	    }
	    /* handled muxed syscalls. */
	    if ((mux_idx = mux_data_index(i)) >= 0) {
		int got_any_before = 0;
		struct mux_data_t* data = &mux_data[mux_idx];

		if (data->before_rules) {
		    int j;
		    for (j=0; j<data->num_muxed; j++) {
	    	        if (data->before_rules[j]) {
	        	    printk(KERN_NOTICE "rules for syscall '%d:%d':\n", i,j);
	        	    printk(KERN_NOTICE " 'before' rules:\n");
	        	    print_tracking_rules(data->before_rules[j]);
	        	    printed = 1;
			    got_any_before = 1;
		        }
		    }
		}
		if (data->after_rules) {
		    int j;
		    for (j=0; j<data->num_muxed; j++) {
	    	        if (data->after_rules[j]) {
			    if (!got_any_before)
	        	        printk(KERN_NOTICE "rules for syscall '%d:%d':\n", i,j);
	        	    printk(KERN_NOTICE " 'after' rules:\n");
	        	    print_tracking_rules(data->after_rules[j]);
	        	    printed = 1;
		        }
		    }
		}
	    }
	}
    }
    if (!printed){
	printk (KERN_NOTICE "no tracking rules registered currently\n");
    }
    printk (KERN_NOTICE "**************************************\n");
    
    up(&tracker_sem);
    
 err_print_module_rules_lock:
    return retval;
}
